﻿using System;
using MarkII.GUI;
using MarkII.Managers;
using MarkII.Misc;
using MarkII.Rotations;
using MarkII.Settings;
using Styx;
using Styx.CommonBot;
using Styx.CommonBot.Routines;
using Styx.TreeSharp;
using Styx.WoWInternals.DBC;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;

namespace MarkII
{
    #region Nested: LocationContextEventArg
    public class WoWContextEventArg : EventArgs
    {
        public readonly WoWContext CurrentWoWContext;
        public readonly WoWContext PreviousWoWContext;

        public WoWContextEventArg(WoWContext currentWoWContext, WoWContext prevWoWContext)
        {
            CurrentWoWContext = currentWoWContext;
            PreviousWoWContext = prevWoWContext;
        }
    }

    #endregion

    public class Main : CombatRoutine
    {
        #region Namings
        private static WoWUnit Pet { get { return StyxWoW.Me.Pet; } }
        private static WoWUnit Focus { get { return StyxWoW.Me.FocusedUnit; } }
        private static WoWUnit Target { get { return StyxWoW.Me.CurrentTarget; } }
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static MainSettings MainSettings { get { return MainSettings.Instance; } }
        #endregion

        #region Delegates
        public delegate T Selection<out T>(object context);
        public delegate WoWPoint LocationRetriever(object context);
        public delegate WoWUnit UnitSelectionDelegate(object context);
        public delegate bool SimpleBooleanDelegate(object context);
        public delegate string SimpleStringDelegate(object context);
        public delegate int SimpleIntDelegate(object context);
        #endregion

        #region Overrides : Combat Routine

        public Main()
        {
            Instance = this;
            BotEvents.OnBotStarted += BotEvents_OnBotStarted;
            BotEvents.Player.OnMapChanged += e =>
            {
                // Don't run this handler if we're not the current routine!
                if (RoutineManager.Current.Name != Name)
                    return;

                //Set Rotation
                RotationSelector.RebuildBehaviors();
                // Only ever update the context. All our internal handlers will use the context changed event
                // so we're not reliant on anything outside of ourselves for updates.
                UpdateContext();
            };
        }

        public static Main Instance { get; private set; }

        /// <summary>
        /// Version information
        /// </summary>
        internal static readonly Version Version = new Version(2, 3, 5);

        /// <summary>
        /// Returns the Combat Routine's name and version
        /// </summary>
        public override string Name
        {
            get { return "The Mark II " + Version; }
        }

        /// <summary>
        /// Combat Routines Class
        /// </summary>
        public override WoWClass Class
        {
            get { return WoWClass.Hunter; }
        }

        /// <summary>
        /// Do we want a GUI or not
        /// </summary>
        public override bool WantButton
        {
            get { return true; }
        }

        /// <summary>
        /// Opens up the GUI on button press
        /// </summary>
        public override void OnButtonPress()
        {
            var gui = new ConfigurationForm();
            var guiSV = new ConfigurationFormSV();
            switch (Me.Specialization)
            {
                case WoWSpec.HunterSurvival:
                    guiSV.ShowDialog();
                    break;
                case WoWSpec.HunterBeastMastery:
                    gui.ShowDialog();
                    break;
            }
        }

        /// <summary>
        /// Called when this CC is selected as the current CC.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();

            Logger.DebugLog("Character Level: {0}", Me.Level);
            Logger.DebugLog("Character Faction: {0}", Me.IsAlliance ? "Alliance" : "Horde");
            Logger.DebugLog("Character Race: {0}", Me.Race);
            Logger.DebugLog("Character Mapname: {0}", Me.MapName);
            Logger.InitLog("LocatonContext: {0}", CurrentWoWContext.ToString());
            Logger.InitLog("Thankyou for choosing Mark II as your Combat Routine");
            Logger.InitLog("Current BotBase is {0}", BotManager.Current.Name);
            Logger.InitLog("Max DPS Setup:");
            Logger.CombatLog("Botbase: Tyrael");
            Logger.CombatLog("TPS: 40");
            Logger.CombatLog("Framelock: On");
            Logger.CombatLog("Plugins: Off");
            Logger.InitLog("Current Routine Version:");
            Logger.CombatLog("- Mark II ~ Hunter v" + Version); // Developer
            Logger.CombatLog("- A Hunter Routine");
            Logger.CombatLog("- by Falldown & Mirabis");
            Logger.InitLog("Your Latency: " + Helpers.MyLatency + " ms");
           // Logger.CombatLog("Remember to set your Custom Lag Tolerance to: 400ms"); BUG: Not needed
           // Logger.CombatLog(" - Accessed in: Game Menu -> Interface -> Combat"); BUG: Not needed
           // Logger.CombatLog("");
            Logger.DebugLog("Press [F9] to calculate Tree Performance (New)");
           // Logger.DebugLog("Press [F6] to show current DPS"); BUG: Not implemented
           // Logger.CombatLog("");

            if (!HotkeyManager.Registered)
            {
                HotkeyManager.RegisterKeys();
            }

            try
            {
                //Get Talents
                TalentManager.Update();
                //Talents for Debugging
                TalentManager.Dumpglyphs();
                //Set Rotation
                RotationSelector.RebuildBehaviors();
                
                // NOTE: Hook these events AFTER the context update.
                OnWoWContextChanged += (orig, ne) =>  Logger.DebugLog(" Context changed, switching rotation to: {0}", CurrentWoWContext.ToString());
            }
            catch (Exception e)
            {
                 Logger.DebugLog(e.ToString());
                BotManager.Current.Stop();
            }

        }

        #region BotEvents_OnBotStarted
        private void BotEvents_OnBotStarted(EventArgs args)
        {
            if (RoutineManager.Current.Name != Name) return;
            switch (BotManager.Current.Name)
            {
                case "Tyrael":
                case "LazyRaider":
                case "Raid Bot":
                case "CombatBot":
                case "Tyrael - BE":
                    break;
                default:
                    TreeRoot.Stop(string.Format("Wrong Botbase! {0} is not supported", BotManager.Current.Name));
                    return;
            }
            DpsMeter.Initialize();
        }

        public override void ShutDown()
        {
            DpsMeter.Shutdown();
            HotkeyManager.RemoveAllKeys();
        }

        #endregion

        #region Pulse
        public override void Pulse()
        { // No pulsing if we're loading or out of the game.
            if (!StyxWoW.IsInGame || !StyxWoW.IsInWorld || !StyxWoW.Me.IsValid)
                return;
            UpdateContext();
            // Double cast shit
            CastManager.PulseDoubleCastEntries();
            /*//Register keys uhm why pulse BUG: Already checked at start of rotation
            if (!HotkeyManager.Registered)
            {
                HotkeyManager.RegisterKeys(); //Hotkeys Registered
            }*/
            PetManager.Pulse();

            if (StyxWoW.Me.CurrentTarget != null)
                DpsMeter.Update();
        }


        #endregion
        #endregion

        #region Overrides : Behavior
        public override Composite CombatBehavior
        {
            get { return RotationSelector.HonorCombatFix(); } //BUG: Temporary Fix
        }

        public override Composite PreCombatBuffBehavior
        {
            get { return RotationSelector.HonorPreCombatFix(); } //BUG: Temporary Fix
        }


        public override Composite RestBehavior
        {
            get { return RotationSelector.HonorPreCombatFix(); } //BUG: Temporary Fix
        }
        #endregion

        #region Hold
        /// <summary>
        /// Holds the rotation for custom actions
        /// </summary>
        /// <returns>RunStatus.Success</returns>
        
        public static Composite Hold()
        {
            return
                new Decorator(
                    ret => Me.HasAnyCachedAura(new[] {"Drink", "Food", "First Aid", "Feign Death"}) || !Unit.ValidUnit(Target, true, true) || Lists.ShouldHold(Target),
                    new Action(delegate { return RunStatus.Success; }));
        }

        #endregion

        #region Custom Units
        private static WoWUnit MDTarget()
        {
            switch (MainSettings.MisdirectOn)
            {
                case MisdirectOn.Me:
                    return StyxWoW.Me;
                case MisdirectOn.Focus:
                    return Unit.ValidFocus(false, true) ? Focus : null;
                case MisdirectOn.Pet:
                    return Me.GotAlivePet ? Pet : null;
                case MisdirectOn.Nobody:
                    return null;

            }
            return null;
        }
        #endregion

        #region LocationContext
        internal static bool ContextEventSubscribed;

        internal static WoWContext LastWoWContext { get; set; }

        public static WoWContext CurrentWoWContext
        {
            get
            {
                Map map = StyxWoW.Me.CurrentMap;

                if (!StyxWoW.IsInGame)
                    return WoWContext.None;

                if ((map.IsBattleground || map.IsArena || MainSettingsPvP.Instance.ForcePvP
                || (MainSettingsPvP.Instance.PlayerPvP && StyxWoW.Me.GotTarget && StyxWoW.Me.CurrentTarget.IsPlayer && Unit.ValidUnit(StyxWoW.Me.CurrentTarget, true, false))) && StyxWoW.Me.Specialization != WoWSpec.HunterSurvival)
                {
                    return WoWContext.PvP;
                }

                return WoWContext.PvE;
            }
        }

        public static event EventHandler<WoWContextEventArg> OnWoWContextChanged;

        public static void UpdateContext()
        {
            // Subscribe to the map change event, so we can automatically update the context.
            if (!ContextEventSubscribed)
            {
                // Subscribe to OnBattlegroundEntered. Just 'cause.
                BotEvents.Battleground.OnBattlegroundEntered += e => UpdateContext();
                ContextEventSubscribed = true;
            }

            WoWContext current = CurrentWoWContext;

            // Can't update the context when it doesn't exist.
            if (current == WoWContext.None)
            {
                 Logger.DebugLog("WowContext doesn't exist.");
                return;
            }

            if (current == LastWoWContext || OnWoWContextChanged == null) return;
            try
            {
                OnWoWContextChanged(Instance, new WoWContextEventArg(current, LastWoWContext));
            }
            catch (Exception e)
            {
                // Eat any exceptions thrown.
                 Logger.DebugLog(e.ToString());
            }
            LastWoWContext = current;
        }
        #endregion

        #region RotationSelector
        public class RotationSelector
        {
            /// <summary>
            /// Until Honorbuddy works again
            /// </summary>
            /// <returns></returns>
            public static Composite HonorPreCombatFix()
            {
                return new PrioritySelector(
                    new Decorator(req => CurrentWoWContext == WoWContext.PvP,
                        new PrioritySelector(
                            new Decorator(req => Me.Specialization == WoWSpec.HunterBeastMastery,
                                BeastMasteryPvP.CreatePrecombat),
                              new Decorator(req => Me.Specialization == WoWSpec.HunterSurvival,
                                SurvivalPvP.CreatePrecombat))),
                    new Decorator(req => CurrentWoWContext == WoWContext.PvE,
                        new PrioritySelector(
                          new Decorator(req => Me.Specialization == WoWSpec.HunterBeastMastery,
                                BeastMasteryPvE.CreatePrecombat),
                          new Decorator(req => Me.Specialization == WoWSpec.HunterSurvival,
                                SurvivalPvP.CreatePrecombat)))
                            
                    );
            }

            /// <summary>
            /// Unit Honorbuddy works again
            /// </summary>
            /// <returns></returns>
            public static Composite HonorCombatFix()
            {
                return new PrioritySelector(
                       new Decorator(req => CurrentWoWContext == WoWContext.PvP,
                        new PrioritySelector(
                            new Decorator(req => Me.Specialization == WoWSpec.HunterBeastMastery,
                                BeastMasteryPvP.CreateCombat),
                              new Decorator(req => Me.Specialization == WoWSpec.HunterSurvival,
                                SurvivalPvP.CreateCombat))),
                    new Decorator(req => CurrentWoWContext == WoWContext.PvE,
                        new PrioritySelector(
                          new Decorator(req => Me.Specialization == WoWSpec.HunterBeastMastery,
                                BeastMasteryPvE.CreateCombat),
                          new Decorator(req => Me.Specialization == WoWSpec.HunterSurvival,
                                SurvivalPvE.CreateCombat)))
                    );
            }

            internal static Composite CreateCombat;
            internal static Composite CreatePreCombat;

            /// <summary>
            /// Creates our behaviors
            /// </summary>
            public static void RebuildBehaviors()
            {
                switch (CurrentWoWContext)
                {
                    #region WoWContext.PvP

                    case WoWContext.PvP:
                        switch (Me.Specialization)
                        {
                            case WoWSpec.HunterBeastMastery:
                                CreateCombat = BeastMasteryPvP.CreateCombat;
                                CreatePreCombat = BeastMasteryPvP.CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvP Specialization - Beast Mastery Activated");
                                break;
                            case WoWSpec.HunterSurvival:
                                CreateCombat = SurvivalPvP.CreateCombat;
                                CreatePreCombat = SurvivalPvP.CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvP Specialization - Survival Activated");
                                break;
                            default:
                                throw new ArgumentOutOfRangeException();
                        }
                        break;

                        #endregion

                        #region WoWContext.PvE

                    case WoWContext.PvE:
                        switch (Me.Specialization)
                        {
                            case WoWSpec.HunterBeastMastery:
                                CreateCombat = BeastMasteryPvE.CreateCombat;
                                CreatePreCombat = BeastMasteryPvE.CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvE Specialization - Beast Mastery Activated");
                                break;
                            case WoWSpec.HunterSurvival:
                                CreateCombat = SurvivalPvE.CreateCombat;
                                CreatePreCombat =  SurvivalPvE.CreatePrecombat;
                                if (CreateCombat != null && CreatePreCombat != null)
                                    Logger.InitLog("PvE Specialization - Survival Activated");
                                break;
                            default:
                                throw new ArgumentOutOfRangeException();
                        }
                        break;
                }

                #endregion
            }
        }
        #endregion
    }
}
/* Mirabises Changes on V2.0.0  
 *  + Cached Auras on all units
 *  + New Cast Method for Beast Mastery to test
 *  + Overal code cleanup
 *  + Fixed broken GUI Settings
 *  + Made some methods to be called as extension
 *  + Removed the huge spell queeing system, to just use latency
 *  + Correct links to thread and change log in gui's now
 *  + Seperated the booleans & Removed redundant specializationchecks
 *  + Seperated
 *  - Broke ur Interface...
 *  - Still have to convert all strings to id's
 *  + AutoPause behavior
 *  + Removed redundant checks
 */

/* Mirabis's Changes on V2.0.5
 *  + BowCanCast Tweaked
 *  - Removed PredicFocus on entire routine and only use it at CancelCast
 *  + Added a new temp fix for rotation changes
 *  + Added The Unit Checks
 * 


/* TO DO List
 * More intelligent Crowd Controlling, especially in arenas
 * Chain more CC's together.  
 * 
 * Smart Concussive Shot (PvP)
 * - Look at our Target and Focus and determine which is a better target for Conc shot.
 * 
 * Add features to Hotkeys
 * - Use cooldowns, Well timed silence, Trap on target location, Use Stun/Daze/Snare. Ask for communitys opinnion.
 * 
 * Add option to use AoE rotation in PvP
 * - Option for Never, Always and Only on NPC's
 * 
 * Improve requirments for all traps.
 * - Especially benefitial for Survival.
 * 
 * Research more spells to dispell with Tranq Shot (HoT's etc), and to Reflect with Deterrence
 * 
 * Improve support for some spells (such as Powershot, Barrage, Wyvern Sting)
 * 
 * Clean out and optimize code
 * - see what needs optimizing. -> Mirabiiiss
 * 
 * Premade presets
 * - Add premade presets
 * 
 * Add support for more pet spells (Hearth of the Phoenix, Bullheaded, Dash)
 */