﻿#region Author Information
// // Leaves - CastManager.cs
// // Created on 30/03/2013 at 01:02 Uhr
// // Author: Mirabis - Moreno Sint Hill
#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using MarkII.Misc;
using MarkII.Settings;
using CommonBehaviors.Actions;
using JetBrains.Annotations;
using Styx;
using Styx.CommonBot;
using Styx.Helpers;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
//using System.Collections.Generic;

namespace MarkII.Managers
{
    public static class CastManager
    {
        //For StopCasting
        public static WoWUnit LastTarget;
        public static string LastCast;

        #region Delegates

        public delegate WoWPoint LocationRetriever(object context);

        public delegate T Selection<out T>(object context);

        #endregion

        /// <summary>
        ///     Indicates if the selected unit has our debuff or not.
        /// </summary>
        /// <param name="onUnit">The specified unit</param>
        /// <param name="auraname">The debuff id to check for</param>
        /// <returns>Returns true, if the the checked aura is present</returns>
        public static bool HasMyAura(this WoWUnit onUnit, string auraname)
        {
            if (onUnit == null) return false;
            WoWAura aura =
                onUnit.GetAllAuras().FirstOrDefault(u => u.CreatorGuid == StyxWoW.Me.Guid && u.Spell.Name == auraname);
            return aura != null;
        }

        #region Casting : Buff
        public static Composite Buff(string name, Main.UnitSelectionDelegate onUnit, Selection<bool> requirements)
        {
            return Buff(name, false, onUnit, requirements);
        }

        public static Composite Buff(string name, bool myBuff, Main.UnitSelectionDelegate onUnit,
                                     Selection<bool> requirements)
        {
            return Buff(name, myBuff, onUnit, requirements, name);
        }

        //private static string _lastBuffCast = string.Empty;
        //private static System.Diagnostics.Stopwatch _castTimer = new System.Diagnostics.Stopwatch();
        /// <summary>
        ///   Creates a behavior to cast a buff by name, with special requirements, on a specific unit. Returns
        ///   RunStatus.Success if successful, RunStatus.Failure otherwise.
        /// </summary>
        /// <remarks>
        ///   Created 5/2/2011.
        /// </remarks>
        /// <param name = "name">The name of the buff</param>
        /// <param name = "myBuff">Check for self debuffs or not</param>
        /// <param name = "onUnit">The on unit</param>
        /// <param name = "requirements">The requirements.</param>
        /// <param name="buffNames"></param>
        /// <returns></returns>
        public static Composite Buff(string name, bool myBuff, Main.UnitSelectionDelegate onUnit,
            Selection<bool> requirements, params string[] buffNames)
        {
            //if (name == _lastBuffCast && _castTimer.IsRunning && _castTimer.ElapsedMilliseconds < 250)
            //{
            //    return new Action(ret => RunStatus.Success);
            //}

            //if (name == _lastBuffCast && StyxWoW.Me.IsCasting)
            //{
            //    _castTimer.Reset();
            //    _castTimer.Start();
            //    return new Action(ret => RunStatus.Success);
            //}

            return
                new Decorator(
                    ret => onUnit != null && onUnit(ret) != null
                    && name != null && !DoubleCastEntries.ContainsKey(name)
                    && buffNames.All(b => myBuff ? !onUnit(ret).HasMyAura(b) : !onUnit(ret).HasAura(b)),
                    new Sequence( // new Action(ctx => _lastBuffCast = name),
                        Cast(name, ret => onUnit(ret), requirements),
                       // new Action(ctx => Logger.CombatLog("Buffing {0} on {1}",name,onUnit(ctx).SafeName())),
                        new DecoratorContinue(ret => GetSpellCastTime(name) > TimeSpan.Zero,
                            new Sequence(new WaitContinue(1, ret => StyxWoW.Me.IsCasting,
                                new Action(ret => UpdateDoubleCastEntries(name,1000)))))));
        }


        /// <summary>
        ///   Creates a behavior to cast a buff by name on yourself. Returns
        ///   RunStatus.Success if successful, RunStatus.Failure otherwise.
        /// </summary>
        /// <remarks>
        ///   Created 5/6/2011.
        /// </remarks>
        /// <param name = "name">The buff name.</param>
        /// <returns>.</returns>
        public static Composite BuffSelf(string name)
        {
            return Buff(name, ret => StyxWoW.Me, ret => true);
        }

        /// <summary>
        ///   Creates a behavior to cast a buff by name on yourself with special requirements. Returns RunStatus.Success if
        ///   successful, RunStatus.Failure otherwise.
        /// </summary>
        /// <remarks>
        ///   Created 5/6/2011.
        /// </remarks>
        /// <param name = "name">The buff name.</param>
        /// <param name = "requirements">The requirements.</param>
        /// <returns>.</returns>
        public static Composite BuffSelf(string name, Selection<bool> requirements)
        {
            return Buff(name, ret => StyxWoW.Me, requirements);
        }

        #endregion

        #region Casting : Cast
        public static Decorator Cast(string spell, Selection<WoWUnit> onUnit, Selection<bool> reqs = null, string reason = "", params object[] args)
        {
            return new Throttle(
                new Decorator(ret => onUnit(ret) != null && BoWCanCast(spell, onUnit(ret), true, true) && (reqs == null || reqs(ret)), 
                new Action(ret =>
                {
                    Logger.CombatLog("F:{0:F1} - Casting {1} on {2} {3}", Helpers.PlayerPower, spell, onUnit(ret).SafeName(), reason);
                    SpellManager.Cast(spell, onUnit(ret));

                })));
        }

        /// <summary>
        /// Auto uses ur Current Target
        /// </summary>
        /// <param name="spell"></param>
        /// <param name="reqs"></param>
        /// <param name="reason"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        public static Decorator Cast(string spell,  Selection<bool> reqs = null, string reason = "", params object[] args)
        {
            return new Throttle(
                new Decorator(ret => StyxWoW.Me.CurrentTarget != null && BoWCanCast(spell, StyxWoW.Me.CurrentTarget,true,true) && (reqs == null || reqs(ret)),
                         new Action(save =>
                             {
                                 Logger.CombatLog("F:{0:F1} - Casting {1} on {2} {3}", Helpers.PlayerPower, spell, StyxWoW.Me.CurrentTarget.SafeName(), reason);
                                 SpellManager.Cast(spell, StyxWoW.Me.CurrentTarget);

                             })));

            // new Action(save => LastCast = spell)//,
            // new Action(save => LastTarget = onUnit(save)))
        }

        //private static WoWUnit CTarget = StyxWoW.Me.CurrentTarget;
        #endregion
        
        #region Casting : Cancel

        public static Composite CancelCast(CanRunDecoratorDelegate cond, RunStatus status, string reason = "No Reason Given", params object[] args)
        {
            return new Decorator(a => IsCasting() && (cond == null || cond(a)),
                new Sequence(
                    new Action(a => Logger.CancelLog("Cancelling {0} - [{1}]", StyxWoW.Me.CastingSpell.Name, reason)),
                    new Action(a => SpellManager.StopCasting()),
                    new Action(delegate { return status; })));
        }

        public static Composite CancelCast(string spell, CanRunDecoratorDelegate cond, RunStatus status, string reason = "No Reason Given", params object[] args)
        {
            return new Decorator(ret => IsCasting() && StyxWoW.Me.CastingSpell.Name == spell && (cond == null || cond(ret)),
                new Sequence(
                    new Action(a => SpellManager.StopCasting()),
                    new Action(a => Logger.CancelLog("Cancelling {0} - [{1}]", spell, reason)),
                    new Action(delegate { return status; })));
        }

        #endregion

        #region Casting : CastOnGround

        public static Composite CastOnGround(string spell, LocationRetriever onLocation)
        {
            return CastOnGround(spell, onLocation, ret => true);
        }

        public static Composite CastOnGround(string spell, LocationRetriever onLocation, CanRunDecoratorDelegate requirements, bool waitForSpell = false)
        {
            return
                new Decorator(ret => requirements(ret) && onLocation != null,
                    new Sequence(
                        new Action(ret => SpellManager.Cast(spell)),
                        new DecoratorContinue(ctx => waitForSpell,
                            new WaitContinue(1, ret => StyxWoW.Me.CurrentPendingCursorSpell != null && StyxWoW.Me.CurrentPendingCursorSpell.Name == spell, new ActionAlwaysSucceed())),
                        new Action(ret => SpellManager.ClickRemoteLocation(onLocation(ret))),
                        new Action(ret => SpellManager.ClickRemoteLocation(onLocation(ret)))));
        }

        #endregion

        #region Extension : Wait


        /// <summary>
        /// Allows waiting for SleepForLagDuration() but ending sooner if condition is met
        /// </summary>
        /// <param name="orUntil">if true will stop waiting sooner than lag maximum</param>
        /// <returns></returns>
        public static Composite CreateWaitForLagDuration(CanRunDecoratorDelegate orUntil = null)
        {
            return new WaitContinue(TimeSpan.FromMilliseconds((StyxWoW.WoWClient.Latency * 2) + 150), orUntil, new ActionAlwaysSucceed());
        }

        public static Composite WaitForGcdOrCastOrChannel()
        {
            return new PrioritySelector(
                WaitForGlobalCooldown(),
                WaitForCast(),
                WaitForChannel()
                );
        }

        private static Composite WaitForGlobalCooldown()
        {
            return new PrioritySelector(
                new Action(ret =>
                {
                    if (IsGlobalCooldown())
                        return RunStatus.Success;

                    return RunStatus.Failure;
                })
                );
        }

        /// <summary>
        ///     Wait for Cast # Mirabis
        /// </summary>
        /// <returns> Runstatus</returns>
        private static Composite WaitForCast()
        {
            return new Action(ret => IsCasting() ? RunStatus.Success : RunStatus.Failure);
        }

        /// <summary>
        ///     Wait for Channel # Mirabis
        /// </summary>
        /// <returns></returns>
        private static Composite WaitForChannel()
        {
            return new Action(ret => IsChannelling() ? RunStatus.Success : RunStatus.Failure);
        }

        /// <summary>
        ///     True/false if questing
        /// </summary>
        /// <returns></returns>
        private static bool IsCasting()
        {
            if (!StyxWoW.Me.IsCasting)
                return false;

            if (StyxWoW.Me.ChannelObjectGuid > 0)
                return false;

            uint latency = StyxWoW.WoWClient.Latency * 2;

            if (StyxWoW.Me.CurrentCastTimeLeft.TotalMilliseconds < latency)
                return false;


            return true;
        }

        /// <summary>
        ///     True/false if Casting
        /// </summary>
        /// <returns></returns>
        private static bool IsChannelling()
        {
            if (!StyxWoW.Me.IsChanneling)
                return false;

            uint latency = StyxWoW.WoWClient.Latency * 2;
            TimeSpan timeLeft = StyxWoW.Me.CurrentChannelTimeLeft;
            return !(timeLeft.TotalMilliseconds < latency);
        }

        /// <summary>
        ///     True/False if GCD is Active
        /// </summary>
        /// <returns></returns>
        public static bool IsGlobalCooldown()
        {
            TimeSpan gcdTimeLeft = SpellManager.GlobalCooldownLeft;
            uint latency = StyxWoW.WoWClient.Latency * 2;
            return gcdTimeLeft.TotalMilliseconds > latency;
        }

        #endregion

        #region Extension : CastTime
        public static TimeSpan GetSpellCastTime(WoWSpell s)
        {
            return TimeSpan.FromMilliseconds(s.CastTime);

        }

        public static TimeSpan GetSpellCastTime(string s)
        {
            SpellFindResults sfr;
            if (SpellManager.FindSpell(s, out sfr))
                return TimeSpan.FromMilliseconds((sfr.Override ?? sfr.Original).CastTime);
            return TimeSpan.Zero;
        }
        #endregion

        #region Extension : CancelAura

        /// <summary>
        ///     Cancel/Remove a aura on x unit
        /// </summary>
        /// <param name="unit"></param>
        /// <param name="aura"></param>
        public static void CancelAura(this WoWUnit unit, string aura)
        {
            WoWAura a = unit.GetAuraByName(aura);
            if (a != null && a.Cancellable)
                a.TryCancelAura();
        }

        #endregion

        #region Extension : Trap Behavior
        public static Composite CreateTrapBehavior(string trapName)
        {
            return CreateTrapBehavior(trapName, ret => StyxWoW.Me.CurrentTarget);
        }

        public static Composite CreateTrapBehavior(string trapName, bool useLauncher)
        {
            return CreateTrapBehavior(trapName, useLauncher, ret => StyxWoW.Me.CurrentTarget);
        }

        public static Composite CreateTrapBehavior(string trapName, Main.UnitSelectionDelegate onUnit)
        {
            return CreateTrapBehavior(trapName, true, onUnit);
        }

        public static Composite CreateTrapBehavior(string trapName, bool useLauncher, Main.UnitSelectionDelegate onUnit, Main.SimpleBooleanDelegate req = null)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => onUnit != null && onUnit(ret) != null
                        && (req == null || req(ret))
                        && onUnit(ret).DistanceSqr < (40 * 40)
                        && SpellManager.HasSpell(trapName) && Helpers.GetSpellCooldown(trapName) == TimeSpan.Zero,
                    new Sequence(
                        new Action(ret => Logger.DebugLog("Trap: use trap launcher requested: {0}", useLauncher)),
                        new PrioritySelector(
                            new Decorator(ret => useLauncher && StyxWoW.Me.HasAura("Trap Launcher"), new ActionAlwaysSucceed()),
                            BuffSelf("Trap Launcher"),
                            new Decorator(ret => !useLauncher, new Action(ret => StyxWoW.Me.CancelAura("Trap Launcher")))
                            ),
                // new Wait(TimeSpan.FromMilliseconds(500), ret => !useLauncher && StyxWoW.Me.HasAura("Trap Launcher"), new ActionAlwaysSucceed()),
                        new Wait(TimeSpan.FromMilliseconds(500),
                            ret => (!useLauncher && !StyxWoW.Me.HasAura("Trap Launcher"))
                                || (useLauncher && StyxWoW.Me.HasAura("Trap Launcher")),
                            new ActionAlwaysSucceed()),
                        new Action(ret => Logger.DebugLog("Trap: launcher aura present = {0}", StyxWoW.Me.HasAura("Trap Launcher"))),
                        new Action(ret => Logger.DebugLog("Trap: cancast = {0}", SpellManager.CanCast(trapName, onUnit(ret)))),

                        new Action(ret => Logger.DebugLog("^{0} trap: {1} on {2}", useLauncher ? "Launch" : "Set", trapName, Helpers.SafeName(onUnit(ret)))),
                // Spell.Cast( trapName, ctx => onUnit(ctx)),
                        new Action(ret => SpellManager.Cast(trapName, onUnit(ret))),

                        CreateWaitForLagDuration(),
                        new Action(ctx => SpellManager.ClickRemoteLocation(onUnit(ctx).Location)),
                        new Action(ret => Logger.DebugLog("Trap: Complete!"))
                        )

                    )
                );
        }

        /*  public static Composite CreateHunterTrapOnAddBehavior(string trapName)
          {
              return new PrioritySelector(
                  ctx => Helpers.NearbyUnfriendlyUnits.OrderBy(u => u.DistanceSqr)
                      .FirstOrDefault(u => u.Combat && u != Target && (!u.IsMoving || u.IsPlayer) && u.DistanceSqr < 40 * 40 && Lists.CCd(u) < 0),
                  new Decorator(
                      ret => ret != null && SpellManager.HasSpell(trapName) && Helpers.GetSpellCooldown(trapName) == TimeSpan.Zero,
                      new Sequence(
                          new Action(ret =>  Logger.DebugLog("AddTrap: make sure we have trap launcher")),
                          new PrioritySelector(
                              Cast("Trap Launcher", ret => StyxWoW.Me, false, false),
                              new ActionAlwaysSucceed()
                              ),
                          new Wait(TimeSpan.FromMilliseconds(500), ret => Me.HasAura("Trap Launcher"), new ActionAlwaysSucceed()),
                          new Action(ret =>  Logger.DebugLog("AddTrap: launcher aura present = {0}", Me.HasAura("Trap Launcher"))),
                          new Action(ret =>  Logger.DebugLog("AddTrap: cancast = {0}", SpellManager.CanCast(trapName, (WoWUnit)ret))),
                  // Spell.Cast( trapName, ctx => (WoWUnit) ctx),
                          new Action(ret =>  Logger.DebugLog("^Launch add trap: {0} on {1}", trapName, Helpers.SafeName((WoWUnit)ret))),
                          new Action(ret => SpellManager.Cast(trapName, (WoWUnit)ret)),
                          Helpers.CreateWaitForLagDuration(),
                          new Action(ctx => SpellManager.ClickRemoteLocation(((WoWUnit)ctx).Location)),
                          new Action(ret =>  Logger.DebugLog("AddTrap: Complete!"))
                          )
                      )
                  );
          }*/
        #endregion

        #region Extension : Prevent Double Cast #Credits to PureRotation
        public static Composite PreventDoubleCast(string spell, double expiryTime, Main.UnitSelectionDelegate onUnit, Main.Selection<bool> reqs = null, string reason = "")
        {
            return new Decorator(
                ret => ((reqs != null && reqs(ret)) || (reqs == null)) && onUnit != null && onUnit(ret) != null && SpellManager.CanCast(spell, onUnit(ret))
                        && !DoubleCastEntries.ContainsKey(spell + onUnit(ret).Guid),
                    new Action(a =>
                        {     
                            if (SpellManager.Cast(spell, onUnit(a)))
                            {
                               Logger.CombatLog("F:{0:F1} - PDCast {1} on {2}, {3}", Helpers.PlayerPower, spell, onUnit(a).SafeName(), reason);
                               UpdateDoubleCastEntries(spell.ToString(CultureInfo.InvariantCulture) + onUnit(a).Guid, expiryTime);
                            }
                        }
                    )      
            );
        }

        static readonly Dictionary<string, DoubleCastSpell> DoubleCastEntries = new Dictionary<string, DoubleCastSpell>();

        private static void UpdateDoubleCastEntries(string spellName, double expiryTime)
        {
            if (DoubleCastEntries.ContainsKey(spellName)) DoubleCastEntries[spellName] = new DoubleCastSpell(spellName, expiryTime, DateTime.UtcNow);
            if (!DoubleCastEntries.ContainsKey(spellName)) DoubleCastEntries.Add(spellName, new DoubleCastSpell(spellName, expiryTime, DateTime.UtcNow));
        }

        internal static void PulseDoubleCastEntries()
        {
            DoubleCastEntries.RemoveAll(t => DateTime.UtcNow.Subtract(t.DoubleCastCurrentTime).TotalSeconds >= t.DoubleCastExpiryTime);
        }

        struct DoubleCastSpell
        {
            public DoubleCastSpell(string spellName, double expiryTime, DateTime currentTime)
                : this()
            {

                DoubleCastSpellName = spellName;
                DoubleCastExpiryTime = expiryTime;
                DoubleCastCurrentTime = currentTime;
            }

            private string DoubleCastSpellName { [UsedImplicitly] get; set; }
            public double DoubleCastExpiryTime { get; private set; }
            public DateTime DoubleCastCurrentTime { get; private set; }
        }


        #endregion

        #region Extension:  CastCobraShot
        /// <summary>
        /// workaround for Spell.Cast("Steady Shot").  fails continuously resulting
        /// in no casts because it never passes the SpellManager.CanCast() test.
        /// Had similar results with traps which are also directly cast because of it
        /// </summary>
        /// <param name="onUnit"></param>
        /// <param name="req"></param>
        /// <param name="reason"></param>
        /// <returns></returns>
        public static Composite CastCobraShot(Main.UnitSelectionDelegate onUnit, Main.SimpleBooleanDelegate req = null, string reason = "")
        {
            return new Decorator(
                ret => onUnit(ret) != null
                    && (req == null || req(ret))
                    && BoWCanCast("Steady Shot", onUnit(ret), true, false),
                    new Throttle(
                new Sequence(
                    new Action(ret => Logger.CombatLog("F:{0:F1} - Casting {1} on {2} - {3}", Helpers.PlayerPower, "Cobra Shot", onUnit(ret).SafeName(), reason)),
                    new Action(ret => SpellManager.Cast("Steady Shot", onUnit(ret)))
                    ))
                );
        }
        #endregion

        #region Extension : BowCanCast
        #region BoWCanCast ( Custom Can Cast by Miracle Business Solutions )
        private static bool BoWCanCast(string name, WoWUnit onUnit, bool rangeCheck, bool checkMovement)
        {
            //Convert "name" to WoWSpell, and check if we have it
            WoWSpell spell;
            if (!SpellManager.Spells.TryGetValue(name, out spell) || onUnit == null)
                //Logger.DebugLog("BowCanCast: I don't have : {0}", name);
                return false;

            //Encounter Specific checks
            if (Main.CurrentWoWContext == WoWContext.PvE && StyxWoW.Me.HasAnyCachedAura(Lists.CannotCast))
                return false;

            //If it's not a Pet Spell, check Line of Spell Sight
            if (!Lists.PetCanCast.Contains(spell.Id) && !onUnit.InLineOfSpellSight)
                return false;

            //If it has a cooldown check if cooldowntimeleft is > then our Latency
            if (!Lists.NoCooldown.Contains(spell.Id) && spell.CooldownTimeLeft.TotalMilliseconds > StyxWoW.WoWClient.Latency * 2) 
                return false;

            if (rangeCheck)
            {
             // If spell has a range specified
                    if (spell.HasRange)
                    {
                      // if unit out of range, false 
                        if (onUnit.Distance > ActualMaxRange(spell, onUnit))
                        {
                            Logger.DebugLog("BowCanCast: cannot cast  {0} @ {1:F1} yds - maximum range is {2:F1}", spell.Name, onUnit.Distance, ActualMaxRange(spell, onUnit));
                            return false;
                        }
                    }
            }
            // if I do not have the focus
            if (Helpers.PlayerPower < PredicCost(spell)) //TODO: Test This
            {
                Logger.DebugLog("BowCanCast:  {0} requires ({1} hfocus / {2} pFocus ) and only {3} available", spell.Name, spell.PowerCost, PredicCost(spell), Helpers.PlayerPower);
                return false;
            }

            if (checkMovement)
            {
                if (spell.CastTime <= 0 || !StyxWoW.Me.IsMoving || Lists.CanCastWhileMoving.Contains(name))
                {
                    //MAD.DiagnosticLog("TreeCanCast: wowSpell {0} is not instant ({1} ms cast time) and we are moving", spell.Name, spell.CastTime);
                    return true;
                }
                Logger.DebugLog("BowCanCast: {0} is not instant ({1} ms cast time) and we are moving", spell.Name, spell.CastTime);
                return false;
            }
            // When not checking movement,  We can cast
            return true;
        }

      #endregion

        #endregion     

        #region Extension : ActualRanges
        /// <summary>
        ///  Returns maximum spell range based on hitbox of unit. 
        /// </summary>
        /// <param name="spell"></param>
        /// <param name="unit"></param>
        /// <returns>Maximum spell range</returns>
        public static float ActualMaxRange(this WoWSpell spell, WoWUnit unit)
        {
            if (spell.MaxRange.Equals(0))
                return 0;
            // 0.3 margin for error
            return unit != null ? spell.MaxRange + unit.CombatReach + 1f : spell.MaxRange;
        }

        public static float ActualMaxRange(string name, WoWUnit unit)
        {
            SpellFindResults sfr;
            if (!SpellManager.FindSpell(name, out sfr))
                return 0f;

            WoWSpell spell = sfr.Override ?? sfr.Original;
            return spell.ActualMaxRange(unit);
        }

        public static float ActualMinRange(this WoWSpell spell, WoWUnit unit)
        {
            if (spell.MinRange.Equals(0))
                return 0;
            // 0.3 margin for error
            return unit != null ? spell.MinRange + unit.CombatReach + 1.6666667f : spell.MinRange;
        }
        #endregion

        #region Extension : Focus Prediction

        /// <summary>
        /// Credits to TuanHa & Edited by Mirabis
        /// </summary>
        /// <returns>The current *Predicted* Focus amount </returns>
        public static double PredicFocus()
        {   
            //BUG :Removed this, only time we should do this is when we cancel Cobra
            if (IsCasting())
            {
                switch (StyxWoW.Me.CastingSpellId)
                {
                    case 77767:
                        return Helpers.PlayerPower + 20;
                    case 56641:
                        return Helpers.PlayerPower + 20;
                    case 19434:
                        return Helpers.PlayerPower - 40;
                }
            }
            return Helpers.PlayerPower;
        }

        /// <summary>
        ///     Predict Spell Cost #Falldown & Mirabis
        /// </summary>
        /// <returns>The Spell Cost</returns>
        public static double PredicCost(WoWSpell spell)
        {
            switch (StyxWoW.Me.Specialization)
            {
                case WoWSpec.HunterSurvival: //No buffs changes focus cost for spells, so return default PowerCost.
                    return spell.PowerCost;
                default:
                    {
                        var powercost = spell.PowerCost;
                        var costsp = SpellManager.Spells["Kill Command"].PowerCost;

                        return StyxWoW.Me.HasCachedAura("The Beast Within") && costsp == 40 ? powercost / 2 : (costsp == 20 ? powercost * 2 : powercost);
                    }
            }
        }
        #endregion

    }
}