using System;
using System.Collections.Generic;
using System.Linq;
using MarkII.Misc;
using Styx;
using Styx.WoWInternals;

namespace MarkII.Managers
{
    internal static class TalentManager
    {
    static TalentManager()
        {
            Talents = new List<Talent>();
            Glyphs = new HashSet<string>();
            Lua.Events.AttachEvent("CHARACTER_POINTS_CHANGED", UpdateTalentManager);
            Lua.Events.AttachEvent("GLYPH_UPDATED", UpdateTalentManager);
            Lua.Events.AttachEvent("ACTIVE_TALENT_GROUP_CHANGED", UpdateTalentManager);
            Lua.Events.AttachEvent("LEARNED_SPELL_IN_TAB", UpdateTalentManager);

        }

        public static WoWSpec CurrentSpec { get; private set; }
        private static List<Talent> Talents { get; set; }
        private static HashSet<string> Glyphs { get; set; }

        public static bool HasGlyph(string glyphName)
        {
            return Glyphs.Count > 0 && Glyphs.Contains(glyphName);
        }

        public static bool HasTalent(int index)
        {
            return Talents.FirstOrDefault(t => t.Index == index).Count != 0;
        }

        public static string Talent90Name()
        {
            if (HasTalent(16))
            {
                return "Glaive Toss";
            }
            else if (HasTalent(17))
            {
                return "Powershot";
            }
            else if (HasTalent(18))
            {
                return "Barrage";
            }
            else return "Not Selected";
        }

        public static string Talent75Name()
        {
            if (TalentManager.HasTalent(13))
            {
                return "A Murder of Crows";
            }
            else if (TalentManager.HasTalent(14))
            {
                return "Blink Strikes";
            }
            else if (TalentManager.HasTalent(15))
            {
                return "Lynx Rush";
            }
            else return "Not Selected";
        }

        public static string Talent60Name()
        {
            if (TalentManager.HasTalent(10))
            {
                return "Fervor";
            }
            else if (TalentManager.HasTalent(11))
            {
                return "Dire Beast";
            }
            else if (TalentManager.HasTalent(12))
            {
                return "Thrill of the Hunt";
            }
            else return "Not Selected";
        }

        public static string Talent45Name()
        {
            if (TalentManager.HasTalent(7))
            {
                return "Exhilaration";
            }
            else if (TalentManager.HasTalent(8))
            {
                return "Aspect of The Iron Hawk";
            }
            else if (TalentManager.HasTalent(9))
            {
                return "Spirit Bond";
            }
            else return "Not Selected";
        }

        public static string Talent30Name()
        {
            if (TalentManager.HasTalent(4))
            {
                return "Silencing Shot";
            }
            else if (TalentManager.HasTalent(5))
            {
                return "Wyvern Sting";
            }
            else if (TalentManager.HasTalent(6))
            {
                return "Intimidation";
            }
            else return "Not Selected";
        }

        public static string Talent15Name()
        {
            if (TalentManager.HasTalent(1))
            {
                return "Posthaste";
            }
            else if (TalentManager.HasTalent(2))
            {
                return "Narrow Escape";
            }
            else if (TalentManager.HasTalent(3))
            {
                return "Crouching Tiger, Hidden Chimera";
            }
            else return "Not Selected";
        }


        private static void UpdateTalentManager(object sender, LuaEventArgs args)
        {
           Update();
           Main.RotationSelector.RebuildBehaviors();
        }

        public static void Update()
        {
            // Don't bother if we're < 10
            if (StyxWoW.Me.Level < 10)
            {
                CurrentSpec = WoWSpec.None;
                return;
            }

            CurrentSpec = StyxWoW.Me.Specialization;

                Talents.Clear();

                // Always 18 talents. 6 rows of 3 talents.
                for (int index = 0; index <= 6 * 3; index++)
                {
                     using (StyxWoW.Memory.AcquireFrame())
                     {
                         var selected = Styx.WoWInternals.Lua.GetReturnVal<int>(string.Format("return GetTalentInfo({0})", index), 4);
                         switch (selected)
                         {
                             case 1:
                                 {
                                     var t = new Talent { Index = index, Count = 1 }; //Name = talentName
                                     Talents.Add(t);
                                 }
                                 break;
                         }
                     }
                    
                }

                Glyphs.Clear();

            using (StyxWoW.Memory.AcquireFrame())
            {
                var glyphCount = Styx.WoWInternals.Lua.GetReturnVal<int>("return GetNumGlyphSockets()", 0);
                
                 Logger.DebugLog("Glyphdetection - GetNumGlyphSockets {0}", glyphCount);

                if (glyphCount != 0)
                {

                    for (int i = 1; i <= glyphCount; i++)
                    {
                        var lua = String.Format("local enabled, glyphType, glyphTooltipIndex, glyphSpellID, icon = GetGlyphSocketInfo({0});if (enabled) then return glyphSpellID else return 0 end", i);
                        var glyphSpellId = Styx.WoWInternals.Lua.GetReturnVal<int>(lua, 0);
                        try
                        {
                            if (glyphSpellId > 0)
                            {
                                 Logger.DebugLog("Glyphdetection - SpellId: {0},Name:{1} ,WoWSpell: {2}", glyphSpellId, WoWSpell.FromId(glyphSpellId).Name, WoWSpell.FromId(glyphSpellId));
                                Glyphs.Add(WoWSpell.FromId(glyphSpellId).Name.Replace("Glyph of ", ""));
                            }
                            else
                            {
                                 Logger.DebugLog("Glyphdetection - Couldn't find all values to detect the Glyph in slot {0}", i);
                            }
                        }
                        catch (Exception ex)
                        {
                             Logger.DebugLog("We couldn't detect your Glyphs");
                             Logger.DebugLog("Report this message to us: " + ex);
                        }
                    }
                }
            }


        }

        public static void Dumpglyphs()
        {
            foreach (var glyph in Glyphs)
            {
                 Logger.DebugLog("{0}", glyph);
            }
        }

        #region Nested type: Talent

        private struct Talent
        {
            public int Count;
            public int Index;
/*
            public string Name;
*/
        }

        #endregion
    }
}
