using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using MarkII.Managers;
using MarkII.Settings;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;
using Action = Styx.TreeSharp.Action;

namespace MarkII.Misc
{
    public static class Helpers
    {
        #region Namings
        
        private static WoWUnit Target { get { return StyxWoW.Me.CurrentTarget; } }
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        #endregion

        #region Get Units & Lists

        public static WoWUnit ArenaPartner;
        public static WoWUnit EnemyPet;
        public static WoWUnit ShamanTotem;
        public static WoWUnit BestDoTClusterUnit;
        public static WoWUnit BestAoEClusterUnit;
        public static WoWUnit MultiDotTarget;

     
        public static Composite AcquireTargets
        {
            get
            {
                return
                    new Action(delegate
                    {
                        //Set Targets
                        ArenaPartner = GetArenaPartner;

                        if (ArenaPartner != null)
                        {
                           // Logger.DebugLog("Arena Partner - {0} {1:F1}% @ {2:F1} yds", SafeName(ArenaPartner), ArenaPartner.HealthPercent, ArenaPartner.Distance);
                        }

                        EnemyPet = GetEnemyPet;

                        ShamanTotem = GetTotems;
                        

                        return RunStatus.Failure;
                    });
            }
        }

        public static Composite AcquirePvETargets
        {
            get
            {
                return
                    new Action(delegate
                    {
                        //Set Targets
                        MultiDotTarget = GetMultiDotTarget;

                        BestAoEClusterUnit = GetBestUnitForCluster(NearbyUnfriendlyUnitsInCombatWithMe.Where(u => Me.IsFacing(u)), 8f);

                        if (Me.Specialization == WoWSpec.HunterSurvival)
                        {
                            BestDoTClusterUnit = GetBestUnitForCluster(NearbyUnfriendlyUnitsInCombatWithMe.Where(u => !u.HasMyCachedAura("Serpent Sting") && Me.IsFacing(u)), 8f);
                        }
                        return RunStatus.Failure;
                    });
            }
        }


        /// <summary>
        /// Takes all the GUID"s of WoWPartyMembers and compares it with a objectManager Query Result.
        /// </summary>
        public static IEnumerable<WoWUnit> PartyMembers
        {
            get
            {
                ulong[] guids = StyxWoW.Me.GroupInfo.RaidMemberGuids
                    .Union(StyxWoW.Me.GroupInfo.PartyMemberGuids)
                    .Union(new[] { StyxWoW.Me.Guid })
                    .Distinct()
                    .ToArray();
                
                return (  // must check inheritance in getobj because of LocalPlayer
                    from p in ObjectManager.GetObjectsOfType<WoWUnit>(true, true)
                    where p.IsFriendly && guids.Any(g => g == p.Guid)
                    select p).ToList();
            }
        }

        /// <summary>
        /// Nearby Valid & Friendly Party Members within 40yd
        /// </summary>
        private static IEnumerable<WoWUnit> NearbyPartyMembers
        {
            get
            {
                return PartyMembers.Where(p => Unit.ValidUnit(p,false,true) &&  p.DistanceSqr <= 40 * 40).ToList();
            }
        }

        public static IEnumerable<WoWUnit> NearbyUnfriendlyUnits
        {
            get { return ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(u => Unit.ValidUnit(u, true, false) && u.DistanceSqr <= 40 * 40).ToList(); }
        }

        public static IEnumerable<WoWUnit> NearbyUnfriendlyPlayers
        {
            get { return ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(p => Unit.ValidUnit(p, true, false) && p.DistanceSqr <= 40 * 40).ToList(); }
        }

        public static IEnumerable<WoWUnit> NearbyFriendlyPlayers
        {
            get { return ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(p => Unit.ValidUnit(p, false, true) && p.DistanceSqr <= 40 * 40).ToList(); }
        }

        public static IEnumerable<WoWUnit> NearbyUnfriendlyUnitsInCombatWithMe
        {
            get { return NearbyUnfriendlyUnits.Where(u => u.IsTargetingUs() && (u.Combat || Lists.TrainingDummies.Contains(u.Entry))); }
        }      

        public static bool UnitsCastingReflectableSpell()
        {
            return NearbyUnfriendlyPlayers.Count(u => ((u.GotTarget && u.CurrentTarget == Me) || (GetUnitCastingReflectableSpell != null && !NearbyFriendlyPlayers.Contains(GetUnitCastingReflectableSpell.CurrentTarget))) && Lists.ReflectCast(u) > 0) > 0;
        }

        public static int DoTUnits()
        {
            return NearbyUnfriendlyUnitsInCombatWithMe.Count(u => u.HasMyCachedAura("Serpent Sting"));
        }

        #region Clustered Units

        public static IEnumerable<WoWUnit> GetRadiusCluster(this WoWUnit target, IEnumerable<WoWUnit> otherUnits, float radius)
        {
            if (target != null)
            {
                var targetLoc = target.Location;
                return otherUnits.Where(u => u.Location.DistanceSqr(targetLoc) <= radius * radius);
            }
            return null;
        }

        // Stolen from Singular with some amendments to tidy code -- Millz
        public static WoWUnit GetBestUnitForCluster(IEnumerable<WoWUnit> units, float clusterRange)
        {
            IEnumerable<WoWUnit> wUnits = units as WoWUnit[] ?? units.ToArray();
            if (units != null && wUnits.Any())
            {
                var firstOrDefault = (from u in wUnits
                                      select new { Count = GetRadiusClusterCount(u, wUnits, clusterRange), Unit = u }).OrderByDescending(a => a.Count).FirstOrDefault();
                if (firstOrDefault != null)
                    return firstOrDefault.Unit;
            }

            return null;
        }

        public static int GetRadiusClusterCount(this WoWUnit target, IEnumerable<WoWUnit> otherUnits, float radius)
        {
            var targetLoc = target.Location;
            return otherUnits.Count(u => u.Location.DistanceSqr(targetLoc) <= radius * radius);
        }

        #endregion

        public static int AddCount(float radius, WoWUnit unit)
        {
            return NearbyUnfriendlyUnitsInCombatWithMe.Count(u => u != null && unit != null && u.Location.Distance2DSqr(unit.Location) < radius * radius) - 1; 
        }

        public static double PartnerPriority(this WoWUnit p)
        {
            double priority = p.HealthPercent;

            if (Lists.CCd(p) > 250)
            { priority -= 20; }

            if (Lists.Snared(p) > 250 || Lists.Silenced(p) > 250 || Lists.Rooted(p) > 250)
            { priority -= 15; }

            if (p.GetClassSpec() == ClassType.Healer)
            { priority -= 30; }

            if (p.GetClassSpec() == ClassType.Tank)
            { priority -= 10; }

            return priority;
        }

        /// <summary>
        /// Get's the Arena Partner by Party Unit Check
        /// </summary>
        public static WoWUnit GetArenaPartner
        {
            get { try {         //Converted to NearbyPartyMembers from Arena Partners
                    return (from unit in NearbyPartyMembers
                            where unit.IsPlayer && !unit.IsMe
                            orderby PartnerPriority(unit) ascending
                            select unit).FirstOrDefault();
                }
                catch (Exception e)
                {
                    Logger.DebugLog("Exception thrown in Helpers.GetArenaPartner: {0}", e);
                    return null;
             } }
        }

        public static WoWUnit GetUnitCastingReflectableSpell
        {
            get
            {
                try
                {
                    return (from unit in NearbyUnfriendlyPlayers
                            where Lists.ReflectCast(unit) > 0
                            orderby unit.CurrentCastTimeLeft descending 
                            select unit).FirstOrDefault();
                }
                catch (Exception e)
                {
                    Logger.DebugLog("Exception thrown in Helpers.GetUnitCastingReflectableSpell: {0}", e);
                    return null;
                }
            }
        }

        public static WoWUnit GetEnemyPet
        {
            get { try {
                return (from unit in NearbyUnfriendlyUnits
                    where unit.IsPet
                    where unit.Combat
                    where unit.IsTargetingUs()
                    where unit.Location.Distance(Me.Pet.Location) <= 20
                    select unit).FirstOrDefault();
                }
                catch (Exception e)
                {
                    Logger.DebugLog("Exception thrown in Helpers.GetEnemyPet: {0}", e);
                    return null;
             }  }
        }

        public static WoWUnit GetMultiDotTarget
        {
            get
            {
                try
                {
                    return (from unit in NearbyUnfriendlyUnits
                            where !unit.IsPet && !unit.IsPlayer
                            where (unit.CurrentHealth > Me.MaxHealth && unit.IsTargetingUs() && unit.Combat) || Lists.TrainingDummies.Contains(unit.Entry)
                            where Me.IsFacing(unit)
                            where unit.HasCachedAuraTimeLeft("Serpent Sting", true).TotalSeconds < 1
                            orderby unit.CurrentHealth descending
                            select unit).FirstOrDefault();
                }
                catch (Exception e)
                {
                    Logger.DebugLog("Exception thrown in Helpers.GetMultiDotTarget: {0}", e);
                    return null;
                }
            }
        }

        public static WoWUnit GetTotems
        {
            get { try {
                return (from unit in NearbyUnfriendlyUnits
                        where
                            (unit.IsTotem &&
                            //Spirit Link Totem 53006
                            //Earthgrab Totem 60561
                            //Earthbind Totem 2630
                            //Windwalk Totem 59717
                            //Mana Tide Totem 10467
                            //Fire Elemental Totem 15439 - Too much health, do not attack
                             (unit.Entry == 53006 ||
                              unit.Entry == 60561 ||
                              unit.Entry == 2630 ||
                              unit.Entry == 59717 ||
                              unit.Entry == 10467)) select unit).FirstOrDefault();
                }
                catch (Exception e)
                {
                    Logger.DebugLog("Exception thrown in Helpers.GetTotems: {0}", e);
                    return null;
            }   }
        }

        #endregion

        #region AutoAttack
        /// <summary>
        ///  Creates a behavior to start auto attacking to current target.
        /// </summary>
        /// <remarks>
        ///  Created 23/05/2011
        /// </remarks>
        /// <param name="includePet"> This will also toggle pet auto attack. </param>
        /// <returns></returns>
        public static Composite CreateAutoAttack(bool includePet)
        {
            const int spellIdAutoShot = 75;

            return new PrioritySelector(
                new Decorator(
                    ret => !StyxWoW.Me.IsAutoAttacking && StyxWoW.Me.AutoRepeatingSpellId != spellIdAutoShot,
                    new Action(ret =>
                    {
                        Lua.DoString("StartAttack()");
                        return RunStatus.Failure;
                    })),
                new Decorator(
                    ret => includePet && StyxWoW.Me.GotAlivePet && (StyxWoW.Me.Pet.CurrentTarget == null || StyxWoW.Me.Pet.CurrentTarget != StyxWoW.Me.CurrentTarget),
                    new Action(
                        delegate
                        {
                            PetManager.CastPetAction("Attack");
                            return RunStatus.Failure;
                        }))
                );
        }

        #endregion
        
        #region PetTargets
        /// <summary>
        /// Checks if pet has this target
        /// </summary>
        /// <param name="u"></param>
        /// <param name="validcheck"></param>
        /// <returns>True/False if Pet has this target</returns>
        public static bool IsPetTarget(this WoWUnit u, bool validcheck)
        {
            if (u == null)
                return false;

            if (validcheck)
            {
                if (!u.IsValid || !u.IsAlive || u.IsDisabled)
                    return false;
            }

                return u.Guid == Me.Pet.CurrentTargetGuid;
        }
        #endregion

        #region LoS Checks

        /// <summary>
        /// Checks if the target is in Line of Sight
        /// 16-3-2013 - Mirabis
        /// </summary>
        /// <param name="from"></param>
        /// <param name="target"></param>
        /// <returns>True/False on Line of Sight</returns>
        public static bool InLoS(this WoWUnit from, WoWUnit target)
        {
            return GameWorld.IsInLineOfSight(@from.Location, target.Location) || target.Location.Distance(Me.Pet.Location) < 6;
        }

        /// <summary>
        /// Checks if target is in line of spell sight
        ///  16-3-2013 - Mirabis
        /// </summary>
        /// <param name="from"></param>
        /// <param name="target"></param>
        /// <returns>True/False on Line of Spell Sight</returns>
        public static bool InLoSS(this WoWUnit from, WoWUnit target)
        {
            return GameWorld.IsInLineOfSpellSight(@from.Location, target.Location) || target.Location.Distance(Me.Pet.Location) < 6;
        }

        public static bool InLoSSS(this WoWUnit from, WoWUnit target)
        {
            return @from.InLoSS(target) || @from.InLoS(target);
        }
        #endregion

        #region SelfControl
        public static bool SelfControl(WoWUnit unit)
        {
            if (unit != null && unit.HasAnyCachedAura(new[] {"Freezing Trap", "Wyvern Sting", "Bad Manner", "Scatter Shot", "Lullaby"}))
                return true;

            else return false;
        }
        #endregion

        #region Cooldowns
        public static TimeSpan GetSpellCooldown(int spell)
        {
            SpellFindResults results;
            if (!SpellOnCooldown(spell)) return TimeSpan.Zero;
            if (SpellManager.FindSpell(spell, out results))
            {
                return results.Override != null ? results.Override.CooldownTimeLeft : results.Original.CooldownTimeLeft;
            }

            return TimeSpan.MaxValue;
        }

        public static TimeSpan GetSpellCooldown(string spell)
        {
            SpellFindResults results;
            if (!SpellOnCooldown(spell)) return TimeSpan.Zero;
            if (SpellManager.FindSpell(spell, out results))
            {
                return results.Override != null ? results.Override.CooldownTimeLeft : results.Original.CooldownTimeLeft;
            }

            return TimeSpan.MaxValue;
        }

        public static bool SpellOnCooldown(string spell)
        {

            SpellFindResults results;
            if (SpellManager.FindSpell(spell, out results))
            {
                return results.Override != null ? results.Override.Cooldown : results.Original.Cooldown;
            }

            return false;
        }

        public static bool SpellOnCooldown(int spell)
        {

            SpellFindResults results;
            if (SpellManager.FindSpell(spell, out results))
            {
                return results.Override != null ? results.Override.Cooldown : results.Original.Cooldown;
            }

            return false;
        }

        public static bool SpellOnGCD(string spell)
        {
            if (GetSpellCooldown(spell).TotalMilliseconds <= SpellManager.GlobalCooldownLeft.TotalMilliseconds + 100)
                return true;

            else return false;
        }

        public static bool IsGlobalCooldown(int time)
        {
            return SpellManager.GlobalCooldownLeft.TotalMilliseconds > time;
        }

        #endregion

        #region Mechanics

        private static readonly List<WoWSpellMechanic> forsakenMechanic = new List<WoWSpellMechanic>
            {
            WoWSpellMechanic.Asleep,
            WoWSpellMechanic.Horrified,
            WoWSpellMechanic.Fleeing,
            WoWSpellMechanic.Charmed
        };

        public static TimeSpan isForsaken(WoWUnit unit)
        {
            var tsTimeRemaining = new TimeSpan(0, 0, 0);
            foreach (WoWAura aura in unit.Auras.Values)
            {
                if (forsakenMechanic.Contains(WoWSpell.FromId(aura.SpellId).Mechanic))
                {
                    if (tsTimeRemaining < aura.TimeLeft)
                        tsTimeRemaining = aura.TimeLeft;
                }
            }
            return tsTimeRemaining;
        }

        #endregion

        #region SafeName

        public static string SafeName(this WoWObject obj)
        {
            if (obj.IsMe)
            {
                return "Myself";
            }

            if (obj.Guid == Me.FocusedUnitGuid)
            {
                return "Focused Unit";
            }

            if (obj.Guid == Me.CurrentTargetGuid)
            {
              return "Current Target";
            }
            if (obj.Guid == Me.Pet.CurrentTargetGuid)
            {
                return "Pet's Target";
            }

            string name;
            var woWPlayer = obj as WoWPlayer;
            if (woWPlayer != null)
            {
                if (RaFHelper.Leader == obj)
                    return "Tank";
                name = woWPlayer.Class.ToString();
            }
            else if (obj is WoWUnit && obj.ToUnit() != null && obj.ToUnit().IsPet)
            {
                name = "Pet";
            }
            else
            {
                name = obj.Name;
            }
            return name;
        }

        #endregion

        #region Latency
        public static uint MyLatency
        {
            get { return StyxWoW.WoWClient.Latency; }
        }
        #endregion

        #region GetAsyncKeyState

        [DllImport("user32.dll")]
        internal static extern short GetAsyncKeyState(Keys vKey);

        public static bool IsKeyDown(Keys key)
        {
            return (GetAsyncKeyState(key)) != 0;
        }

        #endregion

        #region GetTree Performance
        private static readonly Stopwatch TreePerformanceTimer = new Stopwatch(); // lets see if we can get some performance on this one.
        //Credits to Weischbier ***
        internal static Composite TreePerformance(string obj)
        {
            return new Action(ret =>
            {
                if ((Me.Specialization == WoWSpec.HunterBeastMastery && !MainSettings.Instance.TreePerformance) || (Me.Specialization == WoWSpec.HunterSurvival && !MainSettingsSV.Instance.TreePerformance))
                {
                    return RunStatus.Failure;
                }
                if  (GetAsyncKeyState(Keys.F9) != 0)
                {
                
                    if (TreePerformanceTimer.ElapsedMilliseconds > 0)
                    {
                        // NOTE: This dosnt account for Spell casts (meaning the total time is not the time to traverse the tree plus the current cast time of the spell)..this is actual time to traverse the tree.
                        Logger.InitLog("[TreePerformance] Elapsed Time to traverse {0}: {1} ms ({2} ms client lag)", obj,
                                       TreePerformanceTimer.ElapsedMilliseconds, MyLatency);
                        TreePerformanceTimer.Stop();
                        TreePerformanceTimer.Reset();
                    }
                }
                TreePerformanceTimer.Start();


                return RunStatus.Failure;
            });
        }
        #endregion

        #region Overrides of Settings

        public static string SettingsPath(string e)
        {
            return String.Format("{0}\\Settings\\MarkI\\Hotkeys_{1}\\{2}", Utilities.AssemblyDirectory,
                                       StyxWoW.Me.Name, e);
        }
        #endregion

        #region Custom Composites
        internal static Composite FeignDeath()
        {
            return new PrioritySelector(

                   );
        }
        
        #endregion

        #region CallPetNumber

        /// <summary>
        /// Convert Settings enum to int, to call it in cc.
        /// </summary>
        public static int CallPetNumber
        {
            get
            {
                if (Me.Specialization != WoWSpec.HunterSurvival)
                {
                    switch (PetSettings.Instance.CallPet)
                    {
                        case CallPet.Pet1:
                            return 1;
                        case CallPet.Pet2:
                            return 2;
                        case CallPet.Pet3:
                            return 3;
                        case CallPet.Pet4:
                            return 4;
                        case CallPet.Pet5:
                            return 5;
                    }
                    return 0;
                }
                else if (Me.Specialization == WoWSpec.HunterSurvival)
                {
                    switch (PetSettingsSV.Instance.CallPet)
                    {
                        case CallPet.Pet1:
                            return 1;
                        case CallPet.Pet2:
                            return 2;
                        case CallPet.Pet3:
                            return 3;
                        case CallPet.Pet4:
                            return 4;
                        case CallPet.Pet5:
                            return 5;
                    }
                    return 0;
                }
                return 0;
            }
        }

        #endregion

        #region Focus
        /// <summary>
        /// Lua version of Power
        /// </summary>
        public static double PlayerPower
        {
            get
            {
                try
                {
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        return Lua.GetReturnVal<int>("return UnitPower(\"player\");", 0);
                    }
                }
                catch { return 0; }
            }
        }

        #endregion
    }
}
