﻿#region Author Information

// // Mark II - Items.cs
// // Created on 16/04/2013 at 13:42 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using CommonBehaviors.Actions;
using MarkII.Managers;
using MarkII.Settings;
using Styx;
using Styx.CommonBot.Inventory;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.CommonBot;
using Action = Styx.TreeSharp.Action;

namespace MarkII.Misc
{
    public static class Items
    {
        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        private static WoWUnit Target
        {
            get { return StyxWoW.Me.CurrentTarget; }
        }

        /// <summary>
        /// Proffesions,Battlestandard,Healthstones,Potions
        /// </summary>
        public static PrioritySelector UseItems
        {
            get
            {
                return new PrioritySelector(
                    UseEngineerGloves(),
                    CreateBeltBehavior(),
                    CastManager.BuffSelf("Every Man for Himself",
                                    ret =>
                                    Me.Race == WoWRace.Human && 
                                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && CooldownSettings.Instance.Racial) 
                                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && CooldownSettingsPvP.Instance.Racial) 
                                    || (Me.Specialization == WoWSpec.HunterSurvival && CooldownSettingsSV.Instance.Racial))
                                    && Lists.CCd(Me) >= 900),
                    UseTrinkets(),
                    UsePotions()
                    );
            }
        }

        /// <summary>
        /// Uses the item specified (simulated click)
        /// </summary>
        /// <param name="item">The item to use</param>
        private static void UseItem(WoWItem item)
        {
            if (item != null)
            {
                item.Use();
            }
        }

        /// <summary>Creates a behavior to use an equipped item.</summary>
        /// <param name="slot">The slot number of the equipped item. </param>
        /// <returns>success or failure</returns>
        private static Composite UseEquippedItem(uint slot)
        {
            return new PrioritySelector(
                ctx => StyxWoW.Me.Inventory.GetItemBySlot(slot),
                new Decorator(
                    ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                    new Action(ctx => UseItem((WoWItem) ctx))));
        }

        /// <summary>
        ///     Checks for items in the bag, and returns the first item that has an usable spell from the specified string array.
        /// </summary>
        /// <param name="spellNames"> Array of spell names to be check.</param>
        /// <returns></returns>
        private static WoWItem FindFirstUsableItemBySpell(params string[] spellNames)
        {
            List<WoWItem> carried = StyxWoW.Me.CarriedItems;
            // Yes, this is a bit of a hack. But the cost of creating an object each call, is negated by the speed of the Contains from a hash set.
            // So take your optimization bitching elsewhere.
            var spellNameHashes = new HashSet<string>(spellNames);


            return
                carried.Select(i => new {i, spells = i.ItemSpells})
                       .Where(
                           @t =>
                           t.i.ItemInfo != null && t.spells != null && t.spells.Count != 0 && t.i.Usable &&
                           t.i.Cooldown.Equals(0) && t.i.ItemInfo.RequiredLevel <= StyxWoW.Me.Level &&
                           t.spells.Any(
                               s => s.IsValid && s.ActualSpell != null && spellNameHashes.Contains(s.ActualSpell.Name)))
                       .OrderByDescending(@t => t.i.ItemInfo.Level)
                       .Select(t1 => t1.i)
                       .FirstOrDefault();
        }

        private static Composite UseEquippedTrinket(TrinketUsage usage)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => usage == MainSettings.Instance.Trinket1Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket1),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == MainSettings.Instance.Trinket2Usage,
                    new PrioritySelector(
                        ctx => StyxWoW.Me.Inventory.GetItemBySlot((uint) WoWInventorySlot.Trinket2),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    )
                );
        }

        private static Composite UseRefreshment(Refreshment usage)
        {
            return new PrioritySelector(
                new Decorator(
                    ret => usage == Refreshment.Drink,
                    new PrioritySelector(
                        ctx => Consumable.GetBestDrink(true),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == Refreshment.Food,
                    new PrioritySelector(
                        ctx => Consumable.GetBestFood(true),
                        new Decorator(
                            ctx => ctx != null && CanUseEquippedItem((WoWItem) ctx),
                            new Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    ),
                new Decorator(
                    ret => usage == Refreshment.Flask,
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("TODO", "TODO"),
                        new Decorator(
                            ctx =>
                            ctx != null && CanUseEquippedItem((WoWItem) ctx) && !StyxWoW.Me.HasAura(ctx.ToString()),
                            new Action(ctx => UseItem((WoWItem) ctx))
                            )
                        )
                    )
                );
        }

        private static Composite UsePotions()
        {
            return new PrioritySelector(
                new Decorator(ret =>
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && Me.HealthPercent < MainSettings.Instance.SHealthstone)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && Me.HealthPercent < MainSettingsPvP.Instance.SHealthstone)
                    || (Me.Specialization == WoWSpec.HunterSurvival && Me.HealthPercent < MainSettingsSV.Instance.SHealthstone)),

                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Healthstone"),
                        new Decorator(
                            ret => ret != null,
                            new Sequence(
                                new Action(
                                    ctx => Logger.ItemLog("Health Stone used  at {0}", Me.HealthPercent)),
                                new Action(ret => ((WoWItem) ret).UseContainerItem()))))),
                new Decorator(ret =>
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && Me.HealthPercent < MainSettings.Instance.PotionHealth)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && Me.HealthPercent < MainSettingsPvP.Instance.PotionHealth)
                    || (Me.Specialization == WoWSpec.HunterSurvival && Me.HealthPercent < MainSettingsSV.Instance.PotionHealth)),
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Life Spirit", "Healing Potion", "Alchemist's Rejuvenation"),
                        new Decorator(
                            ctx => ctx != null,
                            new Sequence(
                                new Action(
                                    ctx => Logger.ItemLog("Health Potion used at {0}", StyxWoW.Me.HealthPercent)),
                                new Action(ctx => ((WoWItem) ctx).UseContainerItem()))))),
                new Decorator(
                    ret =>
                    Unit.ValidUnit(Target, true, false) && ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && Target.HealthPercent <= MainSettings.Instance.VirmenHealth)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && Target.HealthPercent <= MainSettingsPvP.Instance.VirmenHealth)
                    || (Me.Specialization == WoWSpec.HunterSurvival && Target.HealthPercent <= MainSettingsSV.Instance.VirmenHealth)) 
                    &&
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && MainSettings.Instance.UseVirminPot)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && MainSettingsPvP.Instance.UseVirminPot)
                    || (Me.Specialization == WoWSpec.HunterSurvival && MainSettingsSV.Instance.UseVirminPot)),
                    new PrioritySelector(
                        ctx => FindFirstUsableItemBySpell("Virmen's Bite"),
                        new Decorator(
                            ctx =>
                            ctx != null && (Lists.TargetIsBoss(Target) || Lists.TrainingDummies.Contains(Target.Entry)) &&
                            Me.HasAnyCachedAura(new[] {"Ancient Hysteria", "Time Warp", "Bloodlust", "Heroism"}),
                            new Sequence(
                                new Action(
                                    ctx => Logger.ItemLog("Virmen's Bite used at {0}", StyxWoW.Me.HealthPercent)),
                                new Action(ctx => ((WoWItem) ctx).UseContainerItem())))))
                );
        }

        private static Composite UseEngineerGloves()
        {
            return new PrioritySelector(
                new Decorator(ret => 
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && CooldownSettings.Instance.Gloves)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && CooldownSettingsPvP.Instance.Gloves)
                    || (Me.Specialization == WoWSpec.HunterSurvival && CooldownSettingsSV.Instance.Gloves)), 
                    UseEquippedItem((uint)WoWInventorySlot.Hands))
                );
        }

        private static Composite UseTrinkets()
        {
            return new PrioritySelector(
                new Decorator(ret => 
                    !HotkeyManager.RegenMode, 
                    UseEquippedTrinket(TrinketUsage.Burst)),

                new Decorator(ret =>
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && Me.HealthPercent < MainSettings.Instance.TrinketHealth)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && Me.HealthPercent < MainSettingsPvP.Instance.TrinketHealth)
                    || (Me.Specialization == WoWSpec.HunterSurvival && Me.HealthPercent < MainSettingsSV.Instance.TrinketHealth)),
                    UseEquippedTrinket(TrinketUsage.LowHealth)),

                new Decorator(ret => 
                    ((Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvE && Lists.CCd(Me) > MainSettings.Instance.CCTrinketTime)
                    || (Me.Specialization == WoWSpec.HunterBeastMastery && Main.CurrentWoWContext == WoWContext.PvP && Lists.CCd(Me) > MainSettingsPvP.Instance.CCTrinketTime)
                    || (Me.Specialization == WoWSpec.HunterSurvival && Lists.CCd(Me) > MainSettingsSV.Instance.CCTrinketTime)),
                    UseEquippedTrinket(TrinketUsage.CrowdControlled))
                );
        }


        public static Composite CreateBeltBehavior(bool waitForSpell = false)
        {
            return new Decorator(ret => Target != null && Target.DistanceSqr < (40 * 40) 
                                        && ((Me.Specialization == WoWSpec.HunterBeastMastery && ((Main.CurrentWoWContext == WoWContext.PvE && CooldownSettings.Instance.Belt)
                                        || (Main.CurrentWoWContext == WoWContext.PvP && CooldownSettingsPvP.Instance.Belt)))
                                        || (Me.Specialization == WoWSpec.HunterSurvival && CooldownSettingsSV.Instance.Belt)),
                new Sequence(
                    new Action(ret => UseEquippedItem((uint)WoWInventorySlot.Waist)),
                    new DecoratorContinue(ctx => waitForSpell,
                        new WaitContinue(1, ret => StyxWoW.Me.CurrentPendingCursorSpell != null, new ActionAlwaysSucceed())),
                    new Action(ret => SpellManager.ClickRemoteLocation(Target.Location)),
                    new Action(ret => SpellManager.ClickRemoteLocation(Target.Location))
                )
            );
        }



        private static Composite UseBattleStandard()
        {
            return new PrioritySelector(
                );
        }


        /// <summary>
        /// Checks the equiped item cooldown is 0 and its usable
        /// </summary>
        /// <param name="item">the item to query</param>
        /// <returns>true of false</returns>
        public static bool CanUseEquippedItem(WoWItem item)
        {
            try
            {
                // Check for engineering tinkers!
                using (StyxWoW.Memory.AcquireFrame())
                {
                    var itemSpell = Lua.GetReturnVal<string>("return GetItemSpell(" + item.Entry + ")", 0);
                    if (String.IsNullOrEmpty(itemSpell))
                        return false;

                    return item.Usable && item.Cooldown <= 0;
                }
            }
            catch
            {
                return false;
            }
        }

        #region Tier Piece Armor Detection

        /// <summary>Checks the number of Tier pieces a player is wearing </summary>
        private static int NumTierPieces(int txxItemSetId)
        {
            {
                var
                count = Me.Inventory.Equipped.Hands != null && Me.Inventory.Equipped.Hands.ItemInfo.ItemSetId != 0 && Me.Inventory.Equipped.Hands.ItemInfo.ItemSetId == txxItemSetId ? 1 : 0;
                count += Me.Inventory.Equipped.Legs != null && Me.Inventory.Equipped.Legs.ItemInfo.ItemSetId != 0 && Me.Inventory.Equipped.Legs.ItemInfo.ItemSetId == txxItemSetId ? 1 : 0;
                count += Me.Inventory.Equipped.Chest != null && Me.Inventory.Equipped.Chest.ItemInfo.ItemSetId != 0 && Me.Inventory.Equipped.Chest.ItemInfo.ItemSetId == txxItemSetId ? 1 : 0;
                count += Me.Inventory.Equipped.Shoulder != null && Me.Inventory.Equipped.Shoulder.ItemInfo.ItemSetId != 0 && Me.Inventory.Equipped.Shoulder.ItemInfo.ItemSetId == txxItemSetId ? 1 : 0;
                count += Me.Inventory.Equipped.Head != null && Me.Inventory.Equipped.Head.ItemInfo.ItemSetId != 0 && Me.Inventory.Equipped.Head.ItemInfo.ItemSetId == txxItemSetId ? 1 : 0;
                return count;
            }
        }

        /// <summary>Returns true if player has 2pc Tier set Bonus</summary>
        public static bool Has2PcTeirBonus(int txxItemSetId)
        {
            {
                return NumTierPieces(txxItemSetId) >= 2;
            }
        }

        /// <summary>Returns true if player has 4pc Tier set Bonus</summary>
        public static bool Has4PcTeirBonus(int txxItemSetId)
        {
            {
                return NumTierPieces(txxItemSetId) >= 4;
            }
        }

        #endregion Tier Piece Armor Detection
    }
}