﻿#region Author Information

// // Mark II - Unit.cs
// // Created on 16/04/2013 at 13:44 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using Styx;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace MarkII.Misc
{
    public static class Unit
    {
        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        #region Aura Mechanics
        /// <summary>
        ///     Returns wheter the unit is stunned or not by checking Mechanic, if that fails checks auralist
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static bool IsStunned(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Stunned, WoWSpellMechanic.Incapacitated);
        }

        /// <summary>
        ///     Returns wheter the unit is Rooted or not by checking Mechanic, if that fails checks auralist
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static bool IsRooted(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Rooted, WoWSpellMechanic.Shackled);
        }

        /// <summary>
        ///     Returns wheter the unit is Silenced or not by checking Mechanic, if that fails checks auralist
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static bool IsSilenced(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Interrupted, WoWSpellMechanic.Silenced);
        }

        /// <summary>
        ///     Slow is a general term for any effect (debuff) that lowers the target's attack or movement speed.
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static bool IsSlowed(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Slowed, WoWSpellMechanic.Snared, WoWSpellMechanic.Dazed);
        }

        /// <summary>
        ///     Snare effects are a form of crowd control that reduce a victim's movement speed, preventing him from effectively maneuvering in combat.
        /// </summary>
        /// <param name="unit"></param>
        /// <returns></returns>
        public static bool IsSnared(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Slowed, WoWSpellMechanic.Snared, WoWSpellMechanic.Dazed);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsInvulnerable(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Invulnerable, WoWSpellMechanic.Invulnerable2);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsFeared(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Horrified, WoWSpellMechanic.Fleeing);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsCrowdControlled(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Charmed, WoWSpellMechanic.Banished, WoWSpellMechanic.Asleep,
                                            WoWSpellMechanic.Polymorphed);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsSleeped(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Charmed, WoWSpellMechanic.Asleep, WoWSpellMechanic.Sapped);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsEnraged(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Enraged);
        }

        /// <summary>
        /// </summary>
        /// <param name="unit"></param>
        /// <returns>True/False if Spell Immune</returns>
        public static bool IsPolymorphed(this WoWUnit unit)
        {
            return unit.HasCachedAuraWithMechanic(WoWSpellMechanic.Polymorphed);
        }
        #endregion

        #region IsEnemy

        public static bool IsEnemy(WoWUnit u)
        {
            if (u != null)
            {
                //   if (!Me.IsFFAPvPFlagged)
                if (!StyxWoW.Me.CurrentMap.IsArena && !StyxWoW.Me.CurrentMap.IsBattleground)
                {
                    if (!u.IsFriendly)
                    {
                        return true;
                    }
                    return false;
                }
                //if (!UnitIsInMyPartyOrRaid(target))
                return !IsMyPartyRaidMember(u);
            }
            return false;
        }

        #endregion

        #region ValidUnit

        private static bool ValidUnit(this WoWUnit u)
        {
            return u != null && u.IsValid && !u.IsDisabled && u.IsAlive && u.CanSelect && !u.IsNonCombatPet;
        }

        public static bool ValidUnit(WoWUnit unit, bool enemy, bool friendly)
        {
            if (unit != null && ValidUnit(unit) && Lists.Immune(unit) <= 0)
            {
                if (enemy)
                {
                    return IsEnemy(unit);
                }
                if (friendly)
                {
                    return !IsEnemy(unit);
                }
            }
            return false;
        }

        public static bool ValidFocus(bool enemy, bool friendly)
        {
            var unit = Me.FocusedUnit;

            if (unit != null && ValidUnit(unit) && Lists.Immune(unit) <= 0)
            {
                if (enemy)
                {
                    return IsEnemy(unit);
                }
                if (friendly)
                {
                    return !IsEnemy(unit);
                }
            }
            return false;
        }

        #endregion

        #region GetClassSpec

        public static ClassType GetClassSpec(this WoWUnit target)
        {
            if (target == null)
            {
                return ClassType.Empty;
            }

            switch (target.Class)
            {
                case WoWClass.Monk: //TODO: improve check for monk?
                  return target.MaxMana > 0 ? ClassType.Healer : ClassType.Melee;
                case WoWClass.DeathKnight:
                    return ClassType.Melee;
                case WoWClass.Hunter:
                    return ClassType.PRanged;
                case WoWClass.Mage:
                    return ClassType.MRanged;
                case WoWClass.Paladin:
                    return target.MaxMana >= 80000 ? ClassType.Healer : ClassType.Melee;
                case WoWClass.Priest:
                    return target.Shapeshift == ShapeshiftForm.Shadow ? ClassType.MRanged : ClassType.Healer;
                case WoWClass.Rogue:
                    return ClassType.Melee;
                case WoWClass.Warlock:
                    return ClassType.MRanged;
                case WoWClass.Warrior:
                    return ClassType.Melee;
                case WoWClass.Shaman:
                    if (target.MaxMana < 40000)
                        return ClassType.Melee;
                    if (target.Buffs.ContainsKey("Elemental Oath") &&
                        target.Buffs["Elemental Oath"].CreatorGuid == target.Guid)
                        return ClassType.MRanged;
                    return ClassType.Healer;
                case WoWClass.Druid:
                    if (target.Shapeshift == ShapeshiftForm.Moonkin)
                        return ClassType.MRanged;
                    if ((target.Buffs.ContainsKey("Leader of the Pack") &&
                         target.Buffs["Leader of the Pack"].CreatorGuid == target.Guid) || target.MaxMana < 40000)
                        return ClassType.Melee;
                    return ClassType.Healer;
            }
            return ClassType.Empty;
        }

        #endregion

        #region IsTargetingUs

        /// <summary>
        /// checks if unit is targeting you, your minions, a group member, or group pets
        /// </summary>
        /// <param name="u">unit</param>
        /// <returns>true if targeting your guys, false if not</returns>
        public static bool IsTargetingUs(this WoWUnit u)
        {
            return u.IsTargetingMeOrPet || u.IsTargetingAnyMinion || u.IsTargetingMyPartyMember ||
                   u.IsTargetingMyRaidMember || Lists.TrainingDummies.Contains(u.Entry);
        }

        #endregion

        #region IsMyPartyRaidMember

        /// <summary>
        /// Returns whether the unit is in my party or their owner is.
        /// </summary>
        /// <value>My Party/Raid Members.</value>
        public static bool IsMyPartyRaidMember(this WoWUnit u)
        {
            if (u == null || !u.IsValid)
            {
                return false;
            }
            if (u.IsPlayer)
            {
                var player = u.ToPlayer();
                if (player.IsInMyPartyOrRaid || u.IsMe)
                    return true;
            }
            else if (u.IsPet)
            {
                var owner = u.OwnedByUnit.ToPlayer();
                if (owner.IsInMyPartyOrRaid || u.IsMe)
                    return true;
            }
            return false;
        }

        #endregion
    }
}