﻿#region Author Information

// // Mark II - BeastMasteryPvE.cs
// // Created on 12/04/2013 at 19:05 Uhr
// // Author: Mirabis - Moreno Sint Hill
// 

#endregion

using System;
using System.Linq;
using System.Windows.Forms;
using MarkII.Managers;
using MarkII.Misc;
using MarkII.Settings;
using CommonBehaviors.Actions;
using Styx;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
using CM = MarkII.Managers.CastManager;

namespace MarkII.Rotations
{
    public class BeastMasteryPvE
    {
        #region Priority : Combat Selector

        private static Composite CombatSelector
        {
            get
            {
                return new PrioritySelector(
                    Helpers.TreePerformance("BeastMasteryPvE"),
                    AutoPause(),
                    new Decorator(ret => !HotkeyManager.Registered, new Action(ret => HotkeyManager.RegisterKeys())),
                    new Decorator(ret => MainSettings.Instance.AcquireTargets,
                        new Throttle(1, TimeSpan.FromSeconds(1), Helpers.AcquirePvETargets)
                    ),
                    //SetPetActions, 
                    CM.CancelCast("Cobra Shot",ret => CM.PredicFocus() > 110 & Me.CurrentCastTimeLeft.TotalMilliseconds > 700, RunStatus.Failure, "I Have enough focus"),
                    CM.WaitForGcdOrCastOrChannel(), 
                    new Decorator(req => Unit.ValidUnit(Target, true, false), //Valid Target
                        new PrioritySelector(
                            new Throttle(Items.UseItems),
                            PetManager.CreateHunterCallPetBehavior(),
                            DefensiveCooldowns,
                            CrowdControl(),
                            HotkeyBehavior(),
                            TrapBehavior(when => false),

                            new Decorator(req => ((!Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModifierChoice) || MainSettings.Hotkeys.Instance.CCModifierChoice == Keys.None) && !Helpers.IsKeyDown(MainSettings.Hotkeys.Instance.CCModeKeyChoice)) 
                                                 && Me.CurrentPendingCursorSpell == null,                                               
                                new PrioritySelector(
                                    PetManager.CreateCastPetActionOn("Attack", on => Target),
                                    Interrupt(req => true),
                                    FocusRegen(when => HotkeyManager.RegenMode),
                                    MultiDot(when => HotkeyManager.MultiDotMode && MainSettings.Instance.AcquireTargets && Helpers.DoTUnits() < MainSettings.MultiDoTMax),
                                    new Decorator(when => !HotkeyManager.RegenMode && (!HotkeyManager.MultiDotMode || Helpers.MultiDotTarget == null),
                                        new PrioritySelector(
                                            OffensiveCooldowns(when => true),
                                            AoEClusters(when => MainSettings.AoEClusters && (HotkeyManager.AoEMode || Helpers.AddCount(10f, Helpers.BestAoEClusterUnit) >= MainSettings.ClusterCount)),
                                            AreaofEffect(when => MainSettings.AoE && (HotkeyManager.AoEMode || Helpers.AddCount(10f, Target) >= MainSettings.AddCount)),
                                            SingleTarget
                                        )
                                    )
                                )
                            )
                        )
                    )
                );
            }
        }

        #endregion

        #region Priority : Pre Combat Selector

        private static Composite PreCombatSelector
        {
            get
            {
                return new PrioritySelector(
                    new Decorator(ret => !PetSettings.CombatOnly,
                        PetManager.CreateHunterCallPetBehavior()
                    )
                );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : Defensive Cooldowns

        private static Composite DefensiveCooldowns
        {
            get
            {
                return new PrioritySelector(
                    CM.BuffSelf("Exhilaration", ret => MainSettings.Talent45 && TalentManager.HasTalent(7) && NeedExhilaration()),CM.BuffSelf("Feign Death", ret => NeedFeignDeath()),
                    //CM.BuffSelf("Detterence", ret => ShouldDetterence() ),
                    CM.BuffSelf("Master's Call",ret =>MainSettings.MastersCallPvE && Me.GotAlivePet && Me.IsSnared() || Me.IsRooted()),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Me,ret => Me.HealthPercent < PetSettings.SpiritHealMe &&(PetSettings.SpiritMend == SpiritMend.Me ||PetSettings.SpiritMend == SpiritMend.Me_Focus ||PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Focus,ret => Unit.ValidFocus(false, true) &&Focus.HealthPercent < PetSettings.SpiritHealFocus &&(PetSettings.SpiritMend == SpiritMend.Focus ||PetSettings.SpiritMend == SpiritMend.Me_Focus ||PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    PetManager.CreateCastPetActionOn("Spirit Mend", ret => Pet,ret => Pet != null && Pet.HealthPercent < PetSettings.SpiritHealPet &&(PetSettings.SpiritMend == SpiritMend.Pet ||PetSettings.SpiritMend == SpiritMend.Me_Pet ||PetSettings.SpiritMend == SpiritMend.Me_Focus_Pet)),
                    CM.BuffSelf("Mend Pet",ret =>PetSettings.MendPet && Me.GotAlivePet && Pet.HealthPercent <= PetSettings.PetHealth &&!Pet.HasMyCachedAura("Mend Pet")),
                    PetManager.CreateCastPetAction("Heart of the Phoenix",ret =>PetSettings.RevivePet &&(Me.Pet == null || (Me.Pet != null && !Pet.IsAlive))),
                    CM.BuffSelf("Will of The Forsaken", ret => Helpers.isForsaken(Me).TotalSeconds > 1),
                    CM.BuffSelf("Stoneform",ret =>CooldownSettings.Racial && Me.Race == WoWRace.Dwarf && StyxWoW.Me.GetAllAuras().Any(a =>a.Spell.Mechanic == WoWSpellMechanic.Bleeding ||a.Spell.DispelType == WoWDispelType.Disease ||a.Spell.DispelType == WoWDispelType.Poison)) // Defensive cooldown
                    );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : OffensiveCooldowns

        private static Composite OffensiveCooldowns(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.BuffSelf("Focus Fire", ret =>((NeedFocusFire() && Me.CachedStackCount("Frenzy") >= 5) ||(Me.CachedStackCount("Frenzy") >= 1 && Me.HasCachedAuraTimeLeft("Frenzy", false).TotalSeconds <= 2))),
                    CM.BuffSelf("Aspect of the Hawk", ret =>MainSettings.AspectHawk && !Me.IsMoving && !Me.HasCachedAura(109260) &&!Me.HasCachedAura(13165)),
                    PetManager.CreateCastPetActionOn("Rabid", ret => Me,ret =>PetSettings.Rabid && Lists.TargetIsBoss(Target)),
                    CM.BuffSelf("Berserking", ret => CooldownSettings.Racial && Me.Race == WoWRace.Troll),
                    CM.BuffSelf("Blood Fury", ret => CooldownSettings.Racial && Me.Race == WoWRace.Orc),
                    CM.Buff("Hunter's Mark", ret => Target,ret =>MainSettings.HuntersMark &&(Target.HealthPercent > 20 || Lists.TargetIsBoss(Target))),
                    CM.BuffSelf("Fervor", ret => SpellManager.HasSpell("Fervor") && MainSettings.Talent60_FV &&(Helpers.PlayerPower < 60 ||Helpers.PlayerPower < 40 && Me.HasCachedAura("The Beast Within"))),
                    CM.BuffSelf("Rapid Fire", ret => CooldownSettings.RapidFire && Unit.ValidUnit(Target, true, false) && ((Lists.TargetIsBoss(Target) || Lists.TrainingDummies.Contains(Target.Entry)) && (Helpers.SpellOnCooldown("Bestial Wrath") || !CooldownSettings.BestialWrath) && !Me.HasAnyCachedAura(new[] { "The Beast Within", "Bloodlust", "Heroism", "Time Warp", "Ancient Hysteria" }))),
                    CM.Cast("Stampede", ret => CooldownSettings.Stampede && Lists.TargetIsBoss(Target) &&(Target.CurrentHealth > 1000000 ||Lists.TrainingDummies.Contains(Target.Entry))),
                    CM.Cast("Bestial Wrath", ret => CooldownSettings.BestialWrath && Me.GotAlivePet && Lists.HasteBuffs() < 40 && !Me.HasCachedAura("The Beast Within")),
                    CM.BuffSelf("Readiness", ret => Lists.TargetIsBoss(Target) && NeedReadiness() && CooldownSettings.Readiness),
                    CM.BuffSelf("Lifeblood", ret => CooldownSettings.Lifeblood && Target.MaxHealth > Me.MaxHealth)
                )
            );
        }

        #endregion

        //Not Implemented

        #region Rotation Handler : HotkeyBehavior

        private static Composite HotkeyBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    // TODO: Crowd Control hotkeys for PvE
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : Crowd Control

        private static Composite CrowdControl(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.Cast("Distracting Shot", ret => Target,ret => MainSettings.DistractingShot && Me.GotAlivePet && TalentManager.HasGlyph("Distracting Shot") &&Target.CurrentTargetGuid == Me.Guid),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Focus,ret => PetSettings.Lullaby && Unit.ValidFocus(true, false) && !Focus.HasAnyCachedAura(new[]{"Lullaby", "Scatter Shot"}) && Pet.Location.Distance(Focus.Location) <= 20 && !Focus.IsStunned() && Focus.IsCrowdControlled()),
                    CM.Cast("Misdirection", ret => MDTarget(), ret => !Me.HasCachedAura("Misdirection") && ShouldMisdirect())
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : Interrupt

        private static Composite Interrupt(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.Cast("Scatter Shot", ret => ShouldScatterShot()),
                    CM.Cast("Silencing Shot", ret => TalentManager.HasTalent(4) && MainSettings.Talent30_SS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                    CM.Cast("Wyvern Sting", ret => TalentManager.HasTalent(5) && MainSettings.Talent30_WS && (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200)),
                    CM.Buff("Concussive Shot", ret => Target,ret => MainSettings.ConcussiveShot && !Target.IsSnared() && !Lists.TargetIsBoss(Target) && Target.Distance <= 40),
                    CM.Cast("Intimidation",ret => TalentManager.HasTalent(6) && ShouldIntimidate() && Me.Specialization == WoWSpec.HunterBeastMastery),
                    PetManager.CreateCastPetActionOn("Lullaby", ret => Target, ret => PetSettings.LullabyInt && Pet != null && Lists.TargetIsBoss(Target) && !Target.HasAnyCachedAura(new[]{"Lullaby", "Scatter Shot"})&&((Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds < 27 &&Helpers.GetSpellCooldown("Scatter Shot").TotalSeconds > 1)||(MainSettings.ScatterShot == ScatterShot.Never ||MainSettings.ScatterShot == ScatterShot.Defend))&& Pet.Location.Distance(Target.Location) <= 20)
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : FocusRegen

        private static Composite FocusRegen(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                    CM.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && Helpers.PlayerPower < 70 && !Me.HasCachedAura("The Beast Within")),
                    CM.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                    CM.CastCobraShot(ret => Target, ret => Helpers.PlayerPower < 100),
                    CM.Cast("Arcane Shot", ret => MainSettings.ArcaneShot && Helpers.PlayerPower >= 100)
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : MultiDot

        private static Composite MultiDot(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    new ThrottlePasses(1, TimeSpan.FromMilliseconds(300), RunStatus.Failure, CM.Cast("Serpent Sting", ret => Helpers.MultiDotTarget, ret => Helpers.MultiDotTarget != null, "Multi DoT ON"))
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : TrapBehavior

        private static Composite TrapBehavior(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    new Decorator(ret => Target.Distance > 5 && Target.Distance < 40,
                        new PrioritySelector(
                            CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,ret => MainSettings.STEXPLTrapLauncher),
                            CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,ret => MainSettings.ExplosiveTrapLauncher && Helpers.AddCount(10, Target) > 2),
                            CM.CreateTrapBehavior("Ice Trap", true, ret => Target,ret => MainSettings.IceTrapLauncher),
                            CM.CreateTrapBehavior("Snake Trap", true, ret => Target,ret => MainSettings.SnakeTrapLauncher),
                            CM.CreateTrapBehavior("Freezing Trap", true, ret => Target,ret => MainSettings.FreezingTrapLauncher)
                            )),
                    new Decorator(ret => Target.Distance <= 5,
                        new PrioritySelector(
                            CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,ret => MainSettings.STEXPLTrapLauncher),
                            CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,ret => MainSettings.ExplosiveTrap && Helpers.AddCount(10, Target) > 2),
                            CM.CreateTrapBehavior("Ice Trap", false, ret => Target,ret => MainSettings.IceTrap),
                            CM.CreateTrapBehavior("Snake Trap", false, ret => Target,ret => MainSettings.SnakeTrap),
                            CM.CreateTrapBehavior("Freezing Trap", false, ret => Target,ret => MainSettings.FreezingTrap)))
                    ));
        }

        #endregion

        //Finished

        #region Rotation Handler : Single Target

        private static Composite SingleTarget
        {
            get
            {
                //Single Target Rotation
                return new PrioritySelector(                 
                    CM.Cast("Kill Shot", ret => Target.HealthPercent <= 20),
                    CM.Cast("Kill Command", ret => Me.GotAlivePet),
                    new ThrottlePasses(1, TimeSpan.FromMilliseconds(300), RunStatus.Failure, CM.Cast("Serpent Sting", ret => Target, ret => NeedSerpentSting(Target) && !Target.HasMyCachedAura("Serpent Sting"))),
                   
                    new Decorator(ret => NeedCobraShot() && MainSettings.KillCommand && Me.GotAlivePet,
                        new PrioritySelector(
                            CM.CastCobraShot(ret => Target, ret => !Me.HasCachedAura("The Beast Within") && Helpers.PlayerPower < 70 && Helpers.PlayerPower > 40 && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 1000 && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000, "For KC"),
                            CM.CastCobraShot(ret => Target, ret => (Helpers.PlayerPower < 18 || (!Me.HasCachedAura("The Beast Within") && Helpers.PlayerPower < 38)) && Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < 3000, "For KC")
                        )
                    ),

                    new Decorator(ret => Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds < MainSettings.WaitTime && (Helpers.PlayerPower >= 40 || (Me.HasCachedAura("The Beast Within") && Helpers.PlayerPower >= 20)) && Me.GotAlivePet && MainSettings.KillCommand, new ActionAlwaysSucceed()),
                    new Decorator(ret => Helpers.PlayerPower < 60 && !Me.HasCachedAura("The Beast Within"),
                        new PrioritySelector(
                            CM.CastCobraShot(ret => Target, ret => Helpers.GetSpellCooldown("Bestial Wrath").Seconds < 5 && CooldownSettings.BestialWrath, "For BW"),
                            CM.CastCobraShot(ret => Target, ret => TalentManager.HasTalent(13) && Helpers.GetSpellCooldown("A Murder of Crows").Seconds < 5 && MainSettings.Talent75_AMOC, "For AMOC")
                        )
                    ),
                    CM.Buff("A Murder of Crows", ret => Target, ret => TalentManager.HasTalent(13) && MainSettings.Talent75_AMOC && !Target.HasMyCachedAura("A Murder of Crows") && (Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 5).TotalSeconds > 60) && (!CooldownSettings.BestialWrath || Me.HasCachedAura("The Beast Within") || Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 || !Me.GotAlivePet || Lists.HasteBuffs() >= 40)),
                    CM.Cast("Dire Beast", ret => MainSettings.Talent60_DB && TalentManager.HasTalent(11) && (Lists.TargetIsBoss(Target) || Helpers.PlayerPower < 60 || Target.CurrentHealth > 150000) && !Me.HasCachedAura("The Beast Within")),
                    CM.Cast("Glaive Toss", ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                    CM.Cast("Lynx Rush", ret => Target,ret =>MainSettings.Talent75_LXR && TalentManager.HasTalent(15) && Me.GotAlivePet &&Target.HasCachedAuraTimeLeft("Lynx Rush", true).TotalSeconds < 4 && (Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds >= 10 ||Me.HasCachedAura("The Beast Within")) &&(Lists.TargetIsBoss(Target) || DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 20)),
                    CM.Cast("Powershot", ret => MainSettings.Talent90_PWRS && TalentManager.HasTalent(17) &&Me.HasAnyCachedAura(new[]{"Rapid Fire", "Focus Fire", "Bloodlust", "Heroism", "Ancient Hysteria", "Time Warp"}) &&(Helpers.GetSpellCooldown("Kill Command").TotalMilliseconds > 2500 || !Me.GotAlivePet ||!MainSettings.KillCommand)),
                    CM.Cast("Barrage", ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18) && Lists.TargetIsBoss(Target)),

                    CM.CastCobraShot(ret => Target, ret => Helpers.PlayerPower < 80 && Lists.HasteBuffs() >= 40 && NeedCobraShot(), "High Haste"),

                    CM.CastCobraShot(ret => Target, ret => MainSettings.SerpentSting != SerpentSting.Never && Helpers.PlayerPower < 85 && NeedCobraShot() && Target.HasCachedAuraTimeLeft("Serpent Sting", true).TotalMilliseconds < 4000, "Refresh Serpent Sting"),
                    
                    CM.Cast("Arcane Shot",ret => MainSettings.ArcaneShot && (Helpers.PlayerPower > 62 || (Me.HasCachedAura("The Beast Within") && Helpers.PlayerPower > 42) || Me.HasCachedAura("Thrill of the Hunt") || (TalentManager.HasTalent(10) &&Helpers.GetSpellCooldown("Fervor").TotalMilliseconds < 500) || !Me.GotAlivePet || !MainSettings.KillCommand || !Pet.InLoSSS(Target))),
                    new Decorator(ret => NeedCobraShot(),
                        CM.CastCobraShot(ret => Target,ret => Helpers.PlayerPower < MainSettings.FocusAmount || (Me.HasCachedAura("The Beast Within") && Helpers.PlayerPower < MainSettings.FocusAmount * 0.5), "For Focus"))
                );
            }
        }

        #endregion

        //Finished

        #region Rotation Handler : AoEClusters

        private static Composite AoEClusters(Main.Selection<bool> reqs = null)
        {


            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.CreateTrapBehavior("Explosive Trap", true, ret => Helpers.BestAoEClusterUnit, ret => Helpers.BestAoEClusterUnit.Distance >= 5 && MainSettings.AoEExpTrap),
                    CM.CreateTrapBehavior("Explosive Trap", false, ret => Helpers.BestAoEClusterUnit, ret => Helpers.BestAoEClusterUnit.Distance < 5 && MainSettings.AoEExpTrap),
                    CM.BuffSelf("Focus Fire", ret => (NeedFocusFire() && Pet.CachedStackCount("Frenzy") == 5) || (Pet.CachedStackCount("Frenzy") >= 1 && Pet.HasCachedAuraTimeLeft("Frenzy", false).TotalSeconds < 2)),
                    CM.Cast("Bestial Wrath", ret => CooldownSettings.BestialWrath && Helpers.PlayerPower > 60 && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasCachedAura("The Beast Within")),
                    CM.Cast("Dire Beast", ret => Helpers.BestAoEClusterUnit, ret => MainSettings.AoEDireBeast && Helpers.PlayerPower < 70 && !Me.HasCachedAura("The Beast Within")),
                    CM.Cast("Glaive Toss", ret => Helpers.BestAoEClusterUnit, ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                    CM.Cast("Barrage", ret => Helpers.BestAoEClusterUnit, ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),
                    CM.Cast("Multi-Shot", ret => Helpers.BestAoEClusterUnit, ret => MainSettings.AoEClusters && Helpers.BestAoEClusterUnit != null && (Me.CurrentFocus >= 40 || (Me.HasCachedAura("Thrill of the Hunt") && Me.CurrentFocus >= 20) || (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor"))), "- Better Cluster"),           
                    CM.Buff("Lynx Rush", true, ret => Helpers.BestAoEClusterUnit, ret => MainSettings.AoELynxRush && Me.GotAlivePet && Helpers.PlayerPower < 40 && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasCachedAura("The Beast Within") && DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 15),
                    new Decorator(ret => Helpers.PlayerPower < 20 || (!Me.HasAnyCachedAura(new[] { "Thrill of the Hunt", "The Beast Within" }) && (Helpers.PlayerPower < 40 || (Helpers.PlayerPower < 70 && Pet.HasCachedAuraTimeLeft("Beast Cleave", false).TotalSeconds > 1))),
                        CM.Cast("Steady Shot", ret => Helpers.BestAoEClusterUnit, ret => !TalentManager.HasTalent(10) || Helpers.SpellOnCooldown("Fervor"))
                    )
                )
            );
        }

        #endregion

        //Finished

        #region Rotation Handler : Area of Effect

        private static Composite AreaofEffect(Main.Selection<bool> reqs = null)
        {
            return new Decorator(when => reqs == null || reqs(when),
                new PrioritySelector(
                    CM.CreateTrapBehavior("Explosive Trap", true, ret => Target,ret => Target.Distance >= 5 && MainSettings.AoEExpTrap),
                    CM.CreateTrapBehavior("Explosive Trap", false, ret => Target,ret => Target.Distance < 5 && MainSettings.AoEExpTrap),
                    CM.BuffSelf("Focus Fire",ret => (NeedFocusFire() && Pet.CachedStackCount("Frenzy") == 5) || (Pet.CachedStackCount("Frenzy") >= 1 && Pet.HasCachedAuraTimeLeft("Frenzy", false).TotalSeconds < 2)),
                    CM.Cast("Bestial Wrath",ret => CooldownSettings.BestialWrath && Pet.Location.Distance(Target.Location) <= 25 && !Me.HasCachedAura("The Beast Within")),
                    CM.Cast("Kill Shot",ret => MainSettings.KillShot && Target.HealthPercent < 20 &&Helpers.PlayerPower < 40),
                    CM.Cast("Glaive Toss",ret => MainSettings.Talent90_GT && TalentManager.HasTalent(16)),
                    CM.Cast("Barrage",ret => MainSettings.Talent90_BRG && TalentManager.HasTalent(18)),
                    CM.Cast("Multi-Shot", ret => Helpers.BestAoEClusterUnit, ret => MainSettings.AoEClusters && Helpers.BestAoEClusterUnit != null && (Me.CurrentFocus >= 40 || (Me.HasCachedAura("Thrill of the Hunt") && Me.CurrentFocus >= 20) || (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor"))), "Better Cluster"),
                    CM.Cast("Multi-Shot",ret => MainSettings.MultiShot && (!Pet.HasCachedAura("Beast Cleave") || Pet.HasCachedAuraTimeLeft("Beast Cleave", false).TotalSeconds < 1 || Helpers.PlayerPower >= 70 || Me.HasAnyCachedAura(new[] {"The Beast Within", "Thrill of the Hunt"}) || (TalentManager.HasTalent(10) && !Helpers.SpellOnCooldown("Fervor")))),
                    CM.Cast("Dire Beast", ret => MainSettings.AoEDireBeast && Helpers.PlayerPower < 80 && !Me.HasCachedAura("The Beast Within")),
                    CM.Buff("Lynx Rush", true, ret => Target,ret => MainSettings.AoELynxRush && Me.GotAlivePet && Helpers.PlayerPower < 40 && Pet.Location.Distance(Target.Location) <= 25 &&!Me.HasCachedAura("The Beast Within") &&DpsMeter.GetCombatTimeLeft(Target, 3).TotalSeconds > 15),
                    PetManager.CreateCastPetAction("Froststorm Breath",ret => Me.GotAlivePet && PetSettings.FrostromBreath && Pet.Location.Distance(Target.Location) < 10),
                    PetManager.CreateCastPetAction("Burrow Attack",ret => Me.GotAlivePet && PetSettings.Burrow && Pet.Location.Distance(Target.Location) < 10),
                    new Decorator(
                        ret =>
                        Helpers.PlayerPower < 20 ||(!Me.HasAnyCachedAura(new[] {"Thrill of the Hunt", "The Beast Within"}) &&(Helpers.PlayerPower < 40 ||(Helpers.PlayerPower < 70 &&Pet.HasCachedAuraTimeLeft("Beast Cleave", false).TotalSeconds > 1))),
                        CM.Cast("Steady Shot",ret =>!TalentManager.HasTalent(10) || Helpers.SpellOnCooldown("Fervor")))
                    ));
        }

        #endregion

        #region Ext : 

        public static Composite CreateCombat
        {
            get { return CombatSelector; }
        }

        public static Composite CreatePrecombat
        {
            get { return PreCombatSelector; }
        }

        public static Composite HoldOn(WoWUnit u, bool enemy, bool friendly)
        {
            return new Decorator(ret => !Unit.ValidUnit(u, enemy, friendly) || Lists.ShouldHold(u),
                new ActionAlwaysSucceed()
            );
        }

        public static Composite AutoPause()
        {
            return new Decorator(ret => Me.HasAnyCachedAura(new[] {"Drink", "Food", "First Aid", "Feign Death"}),
                new ActionAlwaysSucceed()
            );
        }

        private static WoWUnit MDTarget()
        {
            switch (MainSettings.MisdirectOn)
            {
                case MisdirectOn.Me:
                    return StyxWoW.Me;
                case MisdirectOn.Focus:
                    return Unit.ValidFocus(false, true) ? Focus : null;
                case MisdirectOn.Pet:
                    return Me.GotAlivePet ? Pet : null;
                case MisdirectOn.Nobody:
                    return null;
            }
            return null;
        }

        #endregion

        #region Tidy : Bools

        private static bool ShouldMisdirect()
        {
            if (Target == null)
                return false;

            switch (MainSettings.MisdirectWhen)
            {
                case MisdirectWhen.Never:
                    return false;
                case MisdirectWhen.InsecurelyTanking:
                    var tankable =
                        Helpers.NearbyUnfriendlyUnitsInCombatWithMe.FirstOrDefault(
                            u => u.Location.Distance2DSqr(Target.Location) < 25*25);
                    return (tankable != null && tankable.ThreatInfo.RawPercent < 90) ||
                           MDTarget().GetThreatInfoFor(Target).RawPercent < 90;
                case MisdirectWhen.Always:
                    return true;
            }
            return false;
        }

        private static bool ShouldIntimidate()
        {
            if (Target == null || !Me.GotAlivePet)
                return false;
            switch (MainSettings.Intimidation)
            {
                case Intimidate.Interrupt:
                    return Pet.Location.Distance(Target.Location) < 9 &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case Intimidate.Defend:
                    return Pet.Location.Distance(Target.Location) < 9 && Target.Distance <= 20 &&
                           Target.CurrentTargetGuid == Me.Guid;
                case Intimidate.Interrupt_Defend:
                    return (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                case Intimidate.Interrupt_Defend_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth &&
                           (Target.Distance <= 20 && Target.CurrentTargetGuid == Me.Guid ||
                            (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200));
                case Intimidate.Interrupt_Health:
                    return Target.HealthPercent < MainSettings.IntimidationHealth &&
                           (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case Intimidate.LowHealth:
                    return Target.HealthPercent < MainSettings.IntimidationHealth;
                case Intimidate.Never:
                    return false;
            }
            return false;
        }

        private static bool ShouldScatterShot()
        {
            if (Target == null)
                return false;
            switch (MainSettings.ScatterShot)
            {
                case ScatterShot.Defend:
                    return Target.CurrentTargetGuid == Me.Guid;
                case ScatterShot.Interrupt:
                    return (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case ScatterShot.Interrupt_Defend:
                    return Target.CurrentTargetGuid == Me.Guid || (Lists.InterruptCastNoChannel(Target) > 200 || Lists.InterruptCastChannel(Target) > 200);
                case ScatterShot.Never:
                    return false;
            }
            return false;
        }

        private static bool NeedReadiness()
        {
            if (Target == null)
                return false;
            return Helpers.GetSpellCooldown("Rapid Fire").TotalSeconds > 30 &&
                   Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 12 &&
                   (Helpers.GetSpellCooldown("Kill Command").TotalSeconds > 2 || !MainSettings.KillCommand) &&
                   ReadinessExtension();
        }

        private static bool NeedCobraShot()
        {
            if (Target == null)
                return false;
            return
                /*(Helpers.GetSpellCooldown("Kill Command").TotalSeconds > Math.Round(((Spell.GetSpellCastTime("Cobra Shot").TotalSeconds / ((100 + Helpers.HasteBuffs()) / 100)) - 0.4), 2) || CLua.PlayerPower < 20 || (!Me.HasCachedAura("The Beast Within") && CLua.PlayerPower < 40))
                    && */ (!TalentManager.HasTalent(10) || (Helpers.GetSpellCooldown("Fervor").TotalMilliseconds >750 && Helpers.GetSpellCooldown("Fervor").TotalSeconds <29));
        }

        private static bool NeedSerpentSting(WoWUnit unit)
        {
            if (Target == null)
                return false;

            switch (MainSettings.SerpentSting)
            {
                case SerpentSting.Always:
                    return true;
                case SerpentSting.Sometimes:
                    return Lists.TargetIsBoss(unit) || DpsMeter.GetCombatTimeLeft(unit, 2).TotalSeconds > 12;
                case SerpentSting.Never:
                    return false;
            }
            return false;
        }

        private static bool ReadinessExtension()
        {
            return (TalentManager.HasTalent(15) && Helpers.GetSpellCooldown("Lynx Rush").TotalSeconds > 10 ||
                    !TalentManager.HasTalent(15))
                   &&
                   (TalentManager.HasTalent(16) && Helpers.GetSpellCooldown("Glaive Toss").TotalSeconds > 4 ||
                    !TalentManager.HasTalent(16))
                   &&
                   (TalentManager.HasTalent(17) && Helpers.GetSpellCooldown("Powershot").TotalSeconds > 5 ||
                    !TalentManager.HasTalent(17))
                   &&
                   (TalentManager.HasTalent(18) && Helpers.GetSpellCooldown("Barrage").TotalSeconds > 6 ||
                    !TalentManager.HasTalent(18))
                   &&
                   (TalentManager.HasTalent(10) && Helpers.GetSpellCooldown("Fervor").TotalSeconds > 10 ||
                    !TalentManager.HasTalent(10))
                   &&
                   (TalentManager.HasTalent(13) && Helpers.GetSpellCooldown("A Murder of Crows").TotalSeconds > 15 ||
                    !TalentManager.HasTalent(13))
                   &&
                   (TalentManager.HasTalent(11) && Helpers.GetSpellCooldown("Dire Beast").TotalSeconds > 8 ||
                    !TalentManager.HasTalent(11));
        }

        private static bool NeedFocusFire()
        {
            if (Target == null)
                return false;
            return Me.GotAlivePet && !Me.HasCachedAura("The Beast Within") &&
                   (Helpers.GetSpellCooldown("Bestial Wrath").TotalSeconds > 9 ||
                    !Helpers.SpellOnCooldown("Bestial Wrath") &&
                    (Target.CurrentHealth <= 200000 ||
                     !Me.HasAnyCachedAura(new[]
                         {"Rapid Fire", "Bestial Wrath", "Heroism", "Bloodlust", "Time Warp", "Ancient Hysteria"})));
        }


        private static bool NeedFeignDeath()
        {
            switch (MainSettings.FeignDeath)
            {
                case FeignDeath.Aggro_Health:
                    return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid &&
                           (Target.IsCasting || Target.Distance < 10) && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.HighThreat:
                    return Target.ThreatInfo.RawPercent > 90;
                case FeignDeath.LowHealth:
                    return Target.CurrentTargetGuid == Me.Guid && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.OnAggro:
                    return Target.ThreatInfo.RawPercent > 90 && Target.CurrentTargetGuid == Me.Guid &&
                           (Target.IsCasting || Target.Distance < 10);
                case FeignDeath.Threat_Health:
                    return Target.ThreatInfo.RawPercent > 90 && Me.HealthPercent < MainSettings.FeignHealth;
                case FeignDeath.Never:
                    return false;
            }
            return false;
        }

        private static bool NeedExhilaration()
        {
            if (Target == null)
                return false;
            return Me.HealthPercent < 68 ||
                   (Me.GotAlivePet && Pet.HealthPercent < 15 && PetManager.CanCastPetAction("Heart of the Phoenix") &&
                    Helpers.GetSpellCooldown("Heart of the Phoenix").TotalSeconds > 10) ||
                   (Me.GotAlivePet && Pet.HealthPercent < 15 && !PetManager.CanCastPetAction("Heart of the Phoenix"));
        }

        #endregion

        #region Tidy: Calls

        private static WoWUnit Pet
        {
            get { return StyxWoW.Me.Pet; }
        }

        private static WoWUnit Focus
        {
            get { return StyxWoW.Me.FocusedUnit; }
        }

        private static WoWUnit Target
        {
            get { return StyxWoW.Me.CurrentTarget; }
        }

        private static LocalPlayer Me
        {
            get { return StyxWoW.Me; }
        }

        private static MainSettings MainSettings
        {
            get { return MainSettings.Instance; }
        }

        private static CooldownSettings CooldownSettings
        {
            get { return CooldownSettings.Instance; }
        }

        private static PetSettings PetSettings
        {
            get { return PetSettings.Instance; }
        }

        #endregion
    }
}