﻿using System.ComponentModel;
using MarkII.Misc;
using Styx;
using Styx.Common;
using Styx.Helpers;

namespace MarkII.Settings
{
    internal class MainSettingsPvP : Styx.Helpers.Settings
    {
        private static MainSettingsPvP _instance;
        public static MainSettingsPvP Instance
        {
            get { return _instance ?? (_instance = new MainSettingsPvP()); }
        }

        public MainSettingsPvP()
            : base(SettingsPath + ".config")
        {
        }

        private static string SettingsPath
        {
            get
            {
                return string.Format("{0}\\Settings\\MarkI\\MarkISettings_MainPvP_{1}", Utilities.AssemblyDirectory,
                                     StyxWoW.Me.Name);
            }
        }

/*
        private static bool IsTrinketUsageWanted(TrinketUsage usage)
        {
            return usage == Instance.Trinket1Usage
                || usage == Instance.Trinket2Usage;
        }
*/

        #region Mechanics

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Acquire Targets")]
        [Description("Acquire Special Targets used in Multi Dotting, for example")]
        public bool AcquireTargets { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Set Focus")]
        [Description("Will Force your old focus'd unit back if lost, even if Auto Focus Mode isn't toggled on.")]
        public bool SetFocus { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Mechanics")]
        [DisplayName("Auto Disable Pet Actions")]
        [Description("Automatically disables pet actions you've chosen to let the CR use.")]
        public bool DisablePetActions { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Arena Buffs")]
        [Description("Automatically buff Aspect of the Hawk and Trap Launcher when in Arena Preparation")]
        public bool AutoBuff { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Kill Totems")]
        [Description("Kill Shaman Totems in Arenas, Communists must die")]
        public bool KillTotems { get; set; }


        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Deterrence Reflect")]
        [Description("Use Deterrence to reflect spells when glyphed.")]
        public bool DeterrenceReflect { get; set; }


        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Mechanics")]
        [DisplayName("Hunter's Mark > 40")]
        [Description("Use Hunter's Mark out of combat on targets that are over 40 yds away")]
        public bool HuntersMarkOOC { get; set; }
        #endregion

        #region Category: Items

        //Trinket Pop Health %
        [Setting]
        [Styx.Helpers.DefaultValue(30)]
        [Category("Items")]
        [DisplayName("Health Trinket")]
        [Description("Less than % Health to use a Health Trinket")]
        public int TrinketHealth { get; set; }

        //Trinket 1
        [Setting]
        [Styx.Helpers.DefaultValue(TrinketUsage.Never)]
        [Category("Items")]
        [DisplayName("Trinket 1")]
        [Description("When to use Trinket 1")]
        public TrinketUsage Trinket1Usage { get; set; }

        //Trinket 2
        [Setting]
        [Styx.Helpers.DefaultValue(TrinketUsage.Never)]
        [Category("Items")]
        [DisplayName("Trinket 2")]
        [Description("When to use Trinket 2")]
        public TrinketUsage Trinket2Usage { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(5400)]
        [Category("Items")]
        [DisplayName("CC Trinket Time")]
        [Description("More than X time left (ms) on a Stun / CC to use Trinket")]
        public int CCTrinketTime { get; set; }

        // Healthstone - Health
        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Items")]
        [DisplayName("Healthstone")]
        [Description("Less than % health for using Healthstone")]
        public int SHealthstone { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(30)]
        [Category("Items")]
        [DisplayName("Potion Health")]
        [Description("Less than % Health to use a Health potion")]
        public int PotionHealth { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(95)]
        [Category("Items")]
        [DisplayName("Virmen Health")]
        [Description("Less than % Health to use a Virmen potion")]
        public int VirmenHealth { get; set; }

        //Health Potion
        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Items")]
        [DisplayName("Virmen Pot")]
        [Description("Enable/Disable use of this pot")]
        public bool UseVirminPot { get; set; }

        // Battlestandard - Health
        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Items")]
        [DisplayName("Battlestandard")]
        [Description("My HP Less than % health to use Horde/Alliance Battle Standard")]
        public int SbStandard { get; set; }
        #endregion

        #region Main


        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Main")]
        [DisplayName("Force PvP")]
        [Description("Forces PvP rotation")]
        public bool ForcePvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Main")]
        [DisplayName("Always PvP against Players")]
        [Description("Forces PvP rotation when targetting an attackable player")]
        public bool PlayerPvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(65)]
        [Category("Main")]
        [DisplayName("Focus Amount")]
        [Description("Less than x Focus to cast Steady/Cobra Shot")]
        public int FocusAmount { get; set; }


        [Setting]
        [Styx.Helpers.DefaultValue(300)]
        [Category("Main")]
        [DisplayName("Wait Time")]
        [Description("How long to delay your shots (in ms) for Kill Command")]
        public int WaitTime { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Main")]
        [DisplayName("TreePerformance")]
        [Description("Enable to measure Tree Performance")]
        public bool TreePerformance { get; set; }

        /*  [Setting]
          [Styx.Helpers.DefaultValue(100)]
          [Category("Main")]
          [DisplayName("Time left on GCD to Cast")]
          [Description("Less than x ms left on Global Cooldown to allow casting")]
          public int GCDTimeLeft { get; set; }

          [Setting]
          [Styx.Helpers.DefaultValue(65)]
          [Category("Main")]
          [DisplayName("Time left on GCD to Traverse")]
          [Description("Less than x ms left on Global Cooldown to traverse through the tree")]
          public int GCDTraverse { get; set; } */
        #endregion

        #region Abilities


        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Abilities")]
        [DisplayName("Exhilaration")]
        [Description("Use Exhilaration")]
        public bool Talent45 { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Abilities")]
        [DisplayName("Fervor")]
        [Description("Use Fervor")]
        public bool Talent60_FV { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Abilities")]
        [DisplayName("Dire Beast")]
        [Description("Use Dire Beast")]
        public bool Talent60_DB { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Abilities")]
        [DisplayName("Aspect of the Hawk")]
        [Description("Force Aspect of the Hawk in combat")]
        public bool AspectHawk { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(PvPFeignDeath.Never)]
        [Category("Abilities")]
        [DisplayName("Feign Death")]
        [Description("When to use Feign Death in PvP.")]
        public PvPFeignDeath PvPFeignDeath { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(80)]
        [Category("Abilities")]
        [DisplayName("Feign Health")]
        [Description("Less than % Health to use Feign Death (PvP)")]
        public int FeignHealthPvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(MastersCall.Never)]
        [Category("Abilities")]
        [DisplayName("Master's Call")]
        [Description("Who to use Master's Call on, in PvP.")]
        public MastersCall MastersCall { get; set; }

        #endregion

        #region Offensive Abilities

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Kill Command")]
        [Description("Use Kill Command")]
        public bool KillCommand { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Kill Shot")]
        [Description("Use Kill Shot")]
        public bool KillShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Arcane Shot")]
        [Description("Use Arcane Shot")]
        public bool ArcaneShot { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(SerpentSting.Sometimes)]
        [Category("Attacks")]
        [DisplayName("Serpent Sting")]
        [Description("When to use Serpent Sting.")]
        public SerpentSting SerpentSting { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("A Murder of Crows")]
        [Description("Use A Murder of Crows")]
        public bool Talent75_AMOC { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Lynx Rush")]
        [Description("Use Lynx Rush")]
        public bool Talent75_LXR { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Glaive Toss")]
        [Description("Use Glaive Toss")]
        public bool Talent90_GT { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Powershot")]
        [Description("Use Powershot")]
        public bool Talent90_PWRS { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Attacks")]
        [DisplayName("Barrage")]
        [Description("Use Barrage")]
        public bool Talent90_BRG { get; set; }

        
        #endregion

        #region Non Damage Abilities

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Controllers")]
        [DisplayName("CC: Freezing Trap")]
        [Description("Crowd Control Focused Unit using Freezing Trap Launcher")]
        public bool CCFreezingTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Controllers")]
        [DisplayName("Hunter's Mark")]
        [Description("Use Hunter's Mark")]
        public bool HuntersMark { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(SilencePvP.Never)]
        [Category("Controllers")]
        [DisplayName("Silencing Shot")]
        [Description("When to use Silencing Shot in PvP.")]
        public SilencePvP SilencePvP { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(SilenceWho.Nobody)]
        [Category("Controllers")]
        [DisplayName("Silence On:")]
        [Description("Who to use Silencing Shot on.")]
        public SilenceWho SilenceWho { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(50)]
        [Category("Controllers")]
        [DisplayName("Silencing Shot Health")]
        [Description("Less than X % health on Unit to use Silencing Shot")]
        public int SilenceHealth { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(ScatterWho.Nobody)]
        [Category("Controllers")]
        [DisplayName("Scatter Shot")]
        [Description("Who to use Scatter Shot on.")]
        public ScatterWho ScatterWho { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(TranqWho.Nobody)]
        [Category("Controllers")]
        [DisplayName("Tranquilizing Shot")]
        [Description("Who to use Tranquilizing Shot on.")]
        public TranqWho TranqWho { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(ConcWho.Nobody)]
        [Category("Controllers")]
        [DisplayName("Concussive Shot")]
        [Description("Who to use Concussive Shot on.")]
        public ConcWho ConcWho { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Controllers")]
        [DisplayName("Widow Venom")]
        [Description("Use Widow Venom")]
        public bool WidowVenom { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(WyvernWho.Nobody)]
        [Category("Controllers")]
        [DisplayName("Wyvern Sting")]
        [Description("Who to use Wyvern Sting on.")]
        public WyvernWho WyvernWho { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(Intimidate.Never)]
        [Category("Controllers")]
        [DisplayName("Intimidation")]
        [Description("When to use Intimidation.")]
        public Intimidate Intimidation { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(40)]
        [Category("Controllers")]
        [DisplayName("Intimidation Health")]
        [Description("Less than % Health on target to use Intimidation")]
        public int IntimidationHealth { get; set; }

        #endregion

        #region Traps

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Ice Trap Launcher")]
        [Description("Use Ice Trap Launcher")]
        public bool IceTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Snake Trap Launcher")]
        [Description("Use Snake Trap Launcher")]
        public bool SnakeTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Freezing Trap Launcher")]
        [Description("Use Freezing Trap Launcher")]
        public bool FreezingTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("Explosive Trap Launcher")]
        [Description("Use Explosive Trap Launcher")]
        public bool ExplosiveTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("ST EXPL Trap Launcher")]
        [Description("Use Single Target Explosive Trap Launcher")]
        public bool STEXPLTrapLauncher { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Ice Trap")]
        [Description("Use Ice Trap")]
        public bool IceTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Snake Trap")]
        [Description("Use Snake Trap")]
        public bool SnakeTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(false)]
        [Category("Traps")]
        [DisplayName("Freezing Trap")]
        [Description("Use Freezing Trap")]
        public bool FreezingTrap { get; set; }

        [Setting]
        [Styx.Helpers.DefaultValue(true)]
        [Category("Traps")]
        [DisplayName("Explosive Trap")]
        [Description("Use Explosive Trap")]
        public bool ExplosiveTrap { get; set; }

        #endregion
    }
}
